from typing import Any

import pandas as pd


def assert_df_eq(
    df: pd.DataFrame,
    data: Any,
    columns: Any = None,
    digits=8,
    check_col_order: bool = False,
    check_order: bool = False,
    check_content: bool = True,
    throw=True,
) -> bool:
    df1 = df
    df2 = (
        data if isinstance(data, pd.DataFrame) else pd.DataFrame(data, columns=columns)
    )
    cols = list(df1.columns)
    try:
        if not check_col_order:
            assert sorted(cols) == sorted(
                df2.columns
            ), f"columns mismatch {sorted(cols)}, {sorted(df2.columns)}"
            df2 = df2[cols]
        else:
            assert cols == list(
                df2.columns
            ), f"columns mismatch {cols}, {list(df2.columns)}"
        assert df1.shape == df2.shape, f"shape mismatch {df1.shape}, {df2.shape}"

        if not check_content:
            return True
        if not check_order:
            df1 = df1.sort_values(cols)
            df2 = df2.sort_values(cols)
        df1 = df1.reset_index(drop=True)
        df2 = df2.reset_index(drop=True)
        pd.testing.assert_frame_equal(
            df1, df2, check_less_precise=digits, check_dtype=False
        )
        return True
    except AssertionError:
        if throw:
            raise
        return False



from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Type, Union

from antlr4.ParserRuleContext import ParserRuleContext
from triad.utils.assertion import assert_or_throw

from _qpd_antlr import QPDParser as qp
from _qpd_antlr import QPDVisitor
from qpd.dataframe import Column, DataFrame, DataFrames
from qpd._parser.utils import (
    ColumnMention,
    ColumnMentions,
    Join,
    VisitorContext,
    is_agg,
    normalize_join_type,
)
from qpd.specs import (
    AggFunctionSpec,
    ArgumentSpec,
    OrderBySpec,
    OrderItemSpec,
    WindowFrameSpec,
    WindowFunctionSpec,
    WindowSpec,
    make_windowframe_spec,
    IsValueSpec,
)
from qpd.workflow import WorkflowDataFrame


class VisitorBase(QPDVisitor, VisitorContext):
    def __init__(self, context: VisitorContext, *args: Any, **kwargs: Any):
        VisitorContext.__init__(self, context, *args, **kwargs)

    def copy(self, tp: Type, *args: Any, **kwargs: Any) -> "VisitorBase":
        return tp(self, *args, **kwargs)

    def not_support(
        self, msg: Union[str, ParserRuleContext]
    ) -> None:  # pragma: no cover
        if isinstance(msg, ParserRuleContext):
            raise NotImplementedError(self.to_str(msg))
        raise NotImplementedError(msg)

    def assert_none(self, *nodes: Any) -> None:  # pragma: no cover
        for node in nodes:
            if isinstance(node, list):
                if len(node) > 0:
                    self.not_support(f"{node} is not empty")
            elif node is not None:
                expr = self.to_str(node)
                self.not_support(f"{expr}")

    def assert_support(
        self, cond, msg: Union[str, ParserRuleContext]
    ) -> None:  # pragma: no cover
        if not cond:
            self.not_support(msg)

    def visitCtes(self, ctx: qp.CtesContext) -> None:
        return

    def visitFunctionName(self, ctx: qp.FunctionNameContext) -> str:
        self.assert_support(ctx.qualifiedName() is not None, ctx)
        return self.to_str(ctx.qualifiedName()).lower()


class PreScan(VisitorBase):
    def __init__(self, context: VisitorContext, *args: Any, **kwargs: Any):
        VisitorContext.__init__(self, context, *args, **kwargs)
        self._in_agg = False
        self._in_window = False
        self._non_agg_mentions = ColumnMentions()
        self._agg_funcs: List[ParserRuleContext] = []
        self._group_by_expressions: List[ParserRuleContext] = []
        self._window_funcs: List[ParserRuleContext] = []
        self._having_expressions: List[ParserRuleContext] = []
        self._not_in_agg_star: Any = None
        self._func_args: List[Tuple[ParserRuleContext, bool, bool]] = []

    def visitStar(self, ctx: qp.StarContext) -> None:
        if self.to_str(ctx) == "*":
            self.set("select_all", True)
        else:
            self.add_column_mentions(ctx)
        if not self._in_agg:
            self._not_in_agg_star = ctx

    def visitColumnReference(self, ctx: qp.ColumnReferenceContext) -> None:
        m = self.add_column_mentions(ctx)
        if not self._in_agg:
            self._non_agg_mentions.add(m)

    def visitDereference(self, ctx: qp.DereferenceContext) -> None:
        m = self.add_column_mentions(ctx)
        if not self._in_agg:
            self._non_agg_mentions.add(m)

    def visitAliasedQuery(self, ctx: qp.QueryContext) -> None:
        return

    def visitFirst(self, ctx: qp.FirstContext) -> None:
        self._handle_func_call("first", ctx)

    def visitLast(self, ctx: qp.LastContext) -> None:
        self._handle_func_call("last", ctx)

    def visitFunctionCall(self, ctx: qp.FunctionCallContext) -> None:
        func = self.visit(ctx.functionName())
        if ctx.windowSpec() is None:
            self._handle_func_call(func, ctx)
        else:
            self.set("has_window_func", True)
            self.assert_support(not self._in_agg and not self._in_window, ctx)
            self._window_funcs.append(ctx)
            self._in_window = True
            self.visitChildren(ctx)
            self._in_window = False
        for a in ctx.argument:
            is_col, is_single = PreFunctionArgument(self).visit_argument(a)
            self._func_args.append((a, is_col, is_single))

    def visitPredicate(self, ctx: qp.PredicateContext):
        self.visitChildren(ctx)
        if ctx.expression() is not None:
            for a in ctx.expression():
                is_col, is_single = PreFunctionArgument(self).visit_argument(a)
                self._func_args.append((a, is_col, is_single))

    def visitWindowRef(self, ctx: qp.WindowRefContext) -> None:  # pragma: no cover
        self.not_support(ctx)

    def visitWindowDef(self, ctx: qp.WindowDefContext) -> None:
        self.set("has_window", True)
        super().visitWindowDef(ctx)

    def visitAggregationClause(self, ctx: qp.AggregationClauseContext):
        for e in ctx.expression():
            self._in_agg = True
            self._group_by_expressions.append(e)
            self.visit(e)
            self._in_agg = False

    def visitRegularQuerySpecification(
        self, ctx: qp.RegularQuerySpecificationContext
    ) -> None:
        self.assert_none(
            ctx.lateralView(),
            ctx.windowClause(),
        )
        if ctx.fromClause() is not None:
            self.copy(PreFrom).visit(ctx.fromClause())
        self.visit(ctx.selectClause())
        if ctx.whereClause() is not None:
            self.visit(ctx.whereClause())
        if ctx.aggregationClause() is not None:
            self.visit(ctx.aggregationClause())
        if ctx.havingClause() is not None:
            self.visit(ctx.havingClause())

        # assert not the case: SELECT *, SUM(x) FROM a
        self.assert_support(
            not self.get("has_agg_func", False) or self._not_in_agg_star is None,
            self._not_in_agg_star,
        )

        # assert not the case: SELECT SUM(x), ROW_NUMBER() OVER() FROM a
        self.assert_support(
            not (
                (self.get("has_agg_func", False) or ctx.aggregationClause() is not None)
                and self.get("has_window_func", False)
            ),
            "can't have both aggregation and window",
        )

        if len(self.joins) > 0:
            # get join details
            joined: Set[str] = set([self.joins[0].df_name])
            for i in range(1, len(self.joins)):
                j = self.joins[i]
                right = j.df_name
                v = self.copy(JoinCriteria)
                v._joined, v._right_name = joined, right
                if j.criteria_context is not None:
                    j.set_conditions(v.visit(j.criteria_context))
                else:
                    j.set_conditions([])
                joined.add(right)

        # handle agg
        self.set("non_agg_mentions", self._non_agg_mentions)
        self.set("agg_funcs", self._agg_funcs)
        self.set("group_by_expressions", self._group_by_expressions)
        self.set("window_funcs", self._window_funcs)
        self.set("func_arg_types", {id(x[0]): (x[1], x[2]) for x in self._func_args})

    def _handle_func_call(self, func: str, ctx: Any) -> None:
        if is_agg(func):
            self.set("has_agg_func", True)
            self.assert_support(not self._in_agg and not self._in_window, ctx)
            self._agg_funcs.append(ctx)
            self._in_agg = True
        self.visitChildren(ctx)
        self._in_agg = False


class RelationVisitor(VisitorBase):
    def __init__(self, context: VisitorContext):
        super().__init__(context)

    def visitTableAlias(self, ctx: qp.TableAliasContext) -> str:
        self.assert_none(ctx.identifierList())
        return self.to_str(ctx.strictIdentifier(), "")

    def visitTableName(self, ctx: qp.TableNameContext) -> Tuple[DataFrame, str]:
        self.assert_none(ctx.sample())
        df_name = self.to_str(ctx.multipartIdentifier(), "")
        assert_or_throw(
            df_name in self.dfs,
            KeyError(f"{df_name} is not found in {list(self.dfs.keys())}"),
        )
        alias = self.visitTableAlias(ctx.tableAlias())
        if alias == "":
            alias = df_name
        return self.dfs[df_name], alias

    def visitAliasedQuery(
        self, ctx: qp.AliasedQueryContext
    ) -> Tuple[DataFrame, str]:  # pragma: no cover
        if ctx.tableAlias().strictIdentifier() is not None:
            name = self.to_str(ctx.tableAlias().strictIdentifier(), "")
        else:
            name = self._obj_to_col_name(self.to_str(ctx.query()))
        v = StatementVisitor(
            VisitorContext(
                sql=self.sql, workflow=self.workflow, dfs=DataFrames(self.dfs)
            )
        )
        df = v.visitQuery(ctx.query())
        return df, name

    def visitAliasedRelation(
        self, ctx: qp.AliasedRelationContext
    ) -> Tuple[DataFrame, str]:
        self.assert_none(ctx.sample())
        df, a = self.copy(RelationVisitor).visit(ctx.relation())
        alias = self.visitTableAlias(ctx.tableAlias())
        return df, alias if alias != "" else a

    def visitRelation(self, ctx: qp.RelationContext) -> Tuple[DataFrame, str]:
        self.assert_none(ctx.joinRelation())
        res = self.visit(ctx.relationPrimary())
        if res is None:  # pragma: no cover
            self.not_support(f"{self.to_str(ctx.relationPrimary())} is not supported")
        return res


class PreFrom(RelationVisitor):
    def __init__(self, context: VisitorContext):
        super().__init__(context)

    def visitFromClause(self, ctx: qp.FromClauseContext) -> None:
        rel = ctx.relation()
        self.assert_support(len(rel) == 1, ctx)
        self.assert_none(ctx.lateralView(), ctx.pivotClause())
        self.visit(rel[0])
        self.set("dfs", DataFrames({x.df_name: x.df for x in self.joins}))

    def visitRelation(self, ctx: qp.RelationContext) -> None:  # type: ignore
        joins = ctx.joinRelation()
        df, name = self.copy(RelationVisitor).visit(ctx.relationPrimary())
        self.add_join_relation(Join(name, df))
        for j in joins:
            self.visit(j)

    def visitJoinRelation(self, ctx: qp.JoinRelationContext) -> None:
        join_type = normalize_join_type(self.to_str(ctx.joinType()))
        natural = ctx.NATURAL() is not None
        df, name = self.copy(RelationVisitor).visit(ctx.right)
        self.add_join_relation(
            Join(
                name,
                df,
                join_type=join_type,
                natural=natural,
                criteria_context=ctx.joinCriteria(),
            )
        )


class ColumnMentionVisitor(VisitorBase):
    def __init__(self, context: VisitorContext):
        super().__init__(context)
        self._mentions = ColumnMentions()

    def get_mentions(self, ctx: Any) -> ColumnMentions:
        self._mentions = ColumnMentions()
        if ctx is not None:
            self.visit(ctx)
        return self._mentions

    def visitStar(self, ctx: qp.StarContext) -> None:  # pragma: no cover
        self.not_support(ctx)
        # self._mentions.add(self.add_column_mentions(ctx))

    def visitColumnReference(self, ctx: qp.ColumnReferenceContext) -> None:
        self._mentions.add(self.add_column_mentions(ctx))

    def visitDereference(self, ctx: qp.DereferenceContext) -> None:
        self._mentions.add(self.add_column_mentions(ctx))


class JoinCriteria(VisitorBase):
    def __init__(self, context: VisitorContext):
        super().__init__(context)

    def visitJoinCriteria(
        self, ctx: qp.JoinCriteriaContext
    ) -> List[List[qp.ValueExpressionContext]]:
        self.assert_none(ctx.identifierList())
        res = self.visit(ctx.booleanExpression())
        self.assert_support(res is not None, ctx)
        return list(res)

    def visitPredicated(
        self, ctx: qp.PredicatedContext
    ) -> Iterable[List[qp.ValueExpressionContext]]:
        self.assert_none(ctx.predicate())
        res = self.visit(ctx.valueExpression())
        self.assert_support(res is not None, ctx)
        return res

    def visitLogicalBinary(
        self, ctx: qp.LogicalBinaryContext
    ) -> Iterable[List[qp.ValueExpressionContext]]:
        self.assert_support(ctx.AND() is not None, ctx)
        for x in self.visit(ctx.left):
            yield x
        for y in self.visit(ctx.right):
            yield y

    def visitComparison(
        self, ctx: qp.ComparisonContext
    ) -> Iterable[List[qp.ValueExpressionContext]]:
        self.assert_support(ctx.comparisonOperator().EQ() is not None, ctx)
        v = self.copy(ColumnMentionVisitor)
        left = v.get_mentions(ctx.left)
        left_ctx = ctx.left
        v = self.copy(ColumnMentionVisitor)
        right = v.get_mentions(ctx.right)
        right_ctx = ctx.right
        yield self._fit(left, left_ctx, right, right_ctx)  # type: ignore

    def _fit(
        self,
        left_mentions: ColumnMentions,
        left_ctx: qp.ValueExpressionContext,
        right_mentions: ColumnMentions,
        right_ctx: qp.ValueExpressionContext,
    ) -> List[qp.ValueExpressionContext]:
        left_set = set(x.df_name for x in left_mentions)
        right_set = set(x.df_name for x in right_mentions)
        joined_set = self._joined
        current_set = set([self._right_name])
        if left_set == current_set and right_set.issubset(joined_set):
            return [right_ctx, left_ctx]
        if right_set == current_set and left_set.issubset(joined_set):
            return [left_ctx, right_ctx]
        raise NotImplementedError(self.to_str(left_ctx), self.to_str(right_ctx))


class PreFunctionArgument(VisitorBase):
    def __init__(self, context: VisitorContext):
        super().__init__(context)
        self._has_star = False
        self._has_func = False
        self._cols = 0
        self._constants = 0

    def visitStar(self, ctx: qp.StarContext) -> None:
        self._cols += 1
        self._has_star = True

    def visitColumnReference(self, ctx: qp.ColumnReferenceContext) -> None:
        self._cols += 1

    def visitDereference(self, ctx: qp.DereferenceContext) -> None:
        self._cols += 1

    def visitConstantDefault(self, ctx: qp.ConstantDefaultContext) -> None:
        self._constants += 1

    def visitFunctionCall(self, ctx: qp.FunctionCallContext) -> None:
        self._has_func = True
        super().visitFunctionCall(ctx)

    def visitLogicalNot(self, ctx: qp.LogicalNotContext) -> None:
        self._has_func = True
        super().visitLogicalNot(ctx)

    def visit_argument(self, ctx: Any) -> Tuple[bool, bool]:
        self._has_star = False
        self._has_func = False
        self._cols = 0
        self._constants = 0
        self.visit(ctx)
        is_col = self._cols > 0
        self.assert_support(is_col or self._constants > 0, ctx)  # neither
        if is_col:
            is_single = (
                self._cols == 1
                and not self._has_star
                and not self._has_func
                and self._constants == 0
            )
        else:
            is_single = self._constants == 1 and not self._has_func
        return is_col, is_single


# ----------------------------------------- Visitors for execution --------------------


class StatementVisitor(VisitorBase):
    def __init__(self, context: VisitorContext):
        super().__init__(context)

    def visitRegularQuerySpecification(
        self, ctx: qp.RegularQuerySpecificationContext
    ) -> DataFrame:
        self.copy(PreScan).visit(ctx)
        self.copy(FromVisitor).visit(ctx)
        if ctx.whereClause() is not None:
            self.copy(WhereVisitor).visit(ctx.whereClause())
        if ctx.aggregationClause() is not None or self.get("has_agg_func", False):
            self.copy(AggregationVisitor).visit(ctx)
        elif self.get("has_window_func", False):
            self.copy(WindowVisitor).visit(ctx)
        if ctx.havingClause() is not None:
            self.copy(WhereVisitor).visit(ctx.havingClause())
        self.copy(SelectVisitor).visit(ctx.selectClause())
        return self.current

    def visitSetOperation(self, ctx: qp.SetOperationContext) -> DataFrame:
        op = self.to_str(ctx.operator).lower()
        self.assert_support(op in ["union", "intersect", "except"], op)
        if op == "except":
            op = "except_df"
        unique = (
            ctx.setQuantifier() is None or ctx.setQuantifier().DISTINCT() is not None
        )
        v = StatementVisitor(
            VisitorContext(
                sql=self.sql, workflow=self.workflow, dfs=DataFrames(self.dfs)
            )
        )
        left = v.visit(ctx.left)
        v = StatementVisitor(
            VisitorContext(
                sql=self.sql, workflow=self.workflow, dfs=DataFrames(self.dfs)
            )
        )
        right = v.visit(ctx.right)
        return self.workflow.op_to_df(list(left.keys()), op, left, right, unique=unique)

    def visitQuery(self, ctx: qp.QueryContext) -> DataFrame:
        if ctx.ctes() is not None:
            self.visit(ctx.ctes())
        df = self.visit(ctx.queryTerm())
        return OrganizationVisitor(self).organize(df, ctx.queryOrganization())

    def visitSingleStatement(self, ctx: qp.SingleStatementContext):
        return self.visit(ctx.statement())

    def visitCtes(self, ctx: qp.CtesContext) -> None:
        for c in ctx.namedQuery():
            name, df = self.visitNamedQuery(c)
            self.set("dfs", DataFrames(self.dfs, {name: df}))

    def visitNamedQuery(self, ctx: qp.NamedQueryContext) -> Tuple[str, DataFrame]:
        self.assert_none(ctx.identifierList())
        name = self.to_str(ctx.name, "")
        v = StatementVisitor(
            VisitorContext(
                sql=self.sql, workflow=self.workflow, dfs=DataFrames(self.dfs)
            )
        )
        df = v.visit(ctx.query())
        return name, df


class OrganizationVisitor(VisitorBase):
    def __init__(self, context: VisitorContext):
        super().__init__(context)

    def visitSortItem(self, ctx: qp.SortItemContext) -> OrderItemSpec:
        name = self.to_str(ctx.expression())
        asc = ctx.DESC() is None
        if ctx.FIRST() is None and ctx.LAST() is None:
            na_position = "auto"
        else:
            na_position = "first" if ctx.FIRST() is not None else "last"
        return OrderItemSpec(name, asc, na_position)

    def visitQueryOrganization(
        self, ctx: qp.QueryOrganizationContext
    ) -> Tuple[OrderBySpec, int]:
        self.assert_none(ctx.clusterBy, ctx.distributeBy, ctx.sort, ctx.windowClause())
        if ctx.order is not None:
            items = [self.visitSortItem(x) for x in ctx.order]
            order_by = OrderBySpec(*items)
        else:
            order_by = OrderBySpec()
        if ctx.limit is not None:
            is_col, _ = PreFunctionArgument(self).visit_argument(ctx.limit)
            self.assert_support(not is_col, ctx)
            limit = int(eval(self.to_str(ctx.limit)))
            self.assert_support(limit >= 0, ctx)
        else:
            limit = -1
        return order_by, limit

    def organize(
        self, df: DataFrame, ctx: Optional[qp.QueryOrganizationContext]
    ) -> DataFrame:
        if ctx is None:
            return df
        order_by, limit = self.visitQueryOrganization(ctx)
        if len(order_by) == 0 and limit < 0:
            return df
        return self.workflow.op_to_df(
            list(df.keys()), "order_by_limit", df, order_by, limit
        )


class ExpressionVisitor(VisitorBase):
    def __init__(self, context: VisitorContext):
        super().__init__(context)

    def visitStar(self, ctx: qp.StarContext) -> List[Column]:
        return self._get_columns(ctx)

    def visitColumnReference(self, ctx: qp.ColumnReferenceContext) -> List[Column]:
        return self._get_columns(ctx)

    def visitDereference(self, ctx: qp.DereferenceContext) -> List[Column]:
        return self._get_columns(ctx)

    def visitConstantDefault(self, ctx: qp.ConstantDefaultContext) -> List[Column]:
        raw = self.to_str(ctx, "")
        if raw.lower() == "true":
            value = True
        elif raw.lower() == "false":
            value = False
        else:
            value = eval(raw)
        return [self.workflow.const_to_col(value)]

    def visitParenthesizedExpression(
        self, ctx: qp.ParenthesizedExpressionContext
    ) -> List[Column]:
        return self.visit(ctx.expression())

    def visitArithmeticUnary(  # type: ignore
        self, ctx: qp.ArithmeticUnaryContext
    ) -> List[Column]:
        op = self.to_str(ctx.operator)
        v = self._get_single_column(ctx.valueExpression())
        return [self.workflow.op_to_col("basic_unary_arithmetic_op", v, op)]

    def visitArithmeticBinary(  # type: ignore
        self, ctx: qp.ArithmeticBinaryContext
    ) -> List[Column]:
        op = self.to_str(ctx.operator)
        left, right = self._get_single_left_right(ctx)
        return [self.workflow.op_to_col("binary_arithmetic_op", left, right, op)]

    def visitLogicalNot(self, ctx: qp.LogicalNotContext) -> List[Column]:
        col = self._get_single_column(ctx.booleanExpression())
        return [self.workflow.op_to_col("logical_not", col)]

    def visitPredicated(self, ctx: qp.PredicatedContext) -> List[Column]:
        if ctx.predicate() is None:
            return self.visit(ctx.valueExpression())
        col = self._get_single_column(ctx.valueExpression())
        kind = self.to_str(ctx.predicate().kind).lower()
        positive = ctx.predicate().NOT() is None
        if kind in ["null", "true", "false"]:
            vs = IsValueSpec(kind, positive)
            return [self.workflow.op_to_col("is_value", col, vs)]
        if kind == "in":
            self.assert_support(len(ctx.predicate().expression()) > 0, ctx)
            in_cols: List[Column] = []
            in_values: List[Any] = []
            for e in ctx.predicate().expression():
                is_col, _ = self.get("func_arg_types")[id(e)]
                if not is_col:
                    in_values.append(eval(self.to_str(e)))
                else:
                    in_cols.append(self._get_single_column(e))
            return [
                self.workflow.op_to_col(
                    "is_in", col, *in_cols, *in_values, positive=positive
                )
            ]
        if kind == "between":
            lower = self._get_single_column((ctx.predicate().lower))
            upper = self._get_single_column((ctx.predicate().upper))
            return [
                self.workflow.op_to_col(
                    "is_between", col, lower, upper, positive=positive
                )
            ]
        return self.not_support(ctx)  # type: ignore  # pragma: no cover

    def visitComparison(self, ctx: qp.ComparisonContext) -> List[Column]:
        def to_op(o: qp.ComparisonOperatorContext):
            if o.EQ():
                return "=="
            if o.NEQ() or o.NEQJ():
                return "!="
            if o.LT():
                return "<"
            if o.LTE():
                return "<="
            if o.GT():
                return ">"
            if o.GTE():
                return ">="
            self.not_support("comparator " + self.to_str(o))  # pragma: no cover

        op = to_op(ctx.comparisonOperator())
        left, right = self._get_single_left_right(ctx)
        return [self.workflow.op_to_col("comparison_op", left, right, op)]

    def visitLogicalBinary(self, ctx: qp.LogicalBinaryContext) -> List[Column]:
        op = self.to_str(ctx.operator).lower()
        left, right = self._get_single_left_right(ctx)
        return [self.workflow.op_to_col("binary_logical_op", left, right, op)]

    def visitFirst(self, ctx: qp.FirstContext) -> List[Column]:
        return [self.current[self.get("agg_func_to_col")[id(ctx)]]]

    def visitLast(self, ctx: qp.LastContext) -> List[Column]:
        return [self.current[self.get("agg_func_to_col")[id(ctx)]]]

    def visitFunctionCall(  # type: ignore
        self, ctx: qp.FunctionCallContext
    ) -> List[Column]:
        # this part should be simplified: expresion_to_col
        if ctx.windowSpec() is not None:
            return [self.current[self.get("window_func_to_col")[id(ctx)]]]
        func = self.visit(ctx.functionName())
        if is_agg(func):
            return [self.current[self.get("agg_func_to_col")[id(ctx)]]]
        self.not_support(ctx)  # pragma: no cover

    def visitSearchedCase(self, ctx: qp.SearchedCaseContext) -> List[Column]:
        cols: List[Column] = []
        for when in ctx.whenClause():
            cols += self.visitWhenClause(when)
        cols.append(self._get_single_column(ctx.elseExpression))
        return [self.workflow.op_to_col("case_when", *cols)]

    def visitWhenClause(self, ctx: qp.WhenClauseContext) -> Tuple[Column, Column]:
        return (
            self._get_single_column(ctx.condition),
            self._get_single_column(ctx.result),
        )

    def _get_single_left_right(self, ctx: Any) -> List[Column]:
        left = self._get_single_column(ctx.left)
        right = self._get_single_column(ctx.right)
        return [left, right]

    def _get_single_column(self, ctx: Any) -> Column:
        c = list(self.visit(ctx))
        self.assert_support(len(c) == 1, ctx)
        return c[0]

    def _get_columns(self, ctx: Any) -> List[Column]:
        cols = self.get_column_mentions(ctx)
        return [self.current[c.encoded] for c in cols]


class FromVisitor(ExpressionVisitor):
    def __init__(self, context: VisitorContext):
        super().__init__(context)
        self._encoded_map: Dict[str, str] = {}

    def visitRegularQuerySpecification(
        self, ctx: qp.RegularQuerySpecificationContext
    ) -> None:
        if len(self.joins) == 0:
            self.update_current(None)  # type: ignore
            self.set("encoded_map", self._encoded_map)
            return
        joined = self._extract_df(self.joins[0].df_name)
        for join in self.joins[1:]:
            name = join.df_name
            df = self._extract_df(name)
            left, right = joined, df
            left_cols: List[Column] = []
            right_cols: List[Column] = []
            joined_cols = list(joined.keys())
            if "semi" not in join.join_type and "anti" not in join.join_type:
                joined_cols += list(df.keys())
            on: List[str] = []
            if len(join.conditions) > 0:
                for i in range(len(join.conditions)):
                    join_col_name = f"_{i}_"
                    left_cols.append(
                        self._join_col(join.conditions[i][0], joined).rename(
                            join_col_name
                        )
                    )
                    right_cols.append(
                        self._join_col(join.conditions[i][1], df).rename(join_col_name)
                    )
                    on.append(f"_{i}_")
                left = WorkflowDataFrame(joined, *left_cols)
                right = WorkflowDataFrame(df, *right_cols)
            joined = self.workflow.op_to_df(
                joined_cols, "join", left, right, on=on, join_type=join.join_type
            )
        self.update_current(joined)
        self.set("encoded_map", self._encoded_map)

    def _join_col(self, ctx: Any, df: WorkflowDataFrame) -> Column:
        self.update_current(df)
        return self._get_single_column(ctx)

    def _extract_df(self, name: str) -> WorkflowDataFrame:
        cols: List[Column] = []
        if self.get("select_all", False):
            df = self.dfs[name]
            for k in df.keys():
                cm = ColumnMention(name, k)
                cols.append(df[k].rename(cm.encoded))
                self._encoded_map[cm.encoded] = k
        else:
            for m in self.all_column_mentions:
                if m.df_name == name:
                    cols.append(self.dfs[name][m.col_name].rename(m.encoded))
                    self._encoded_map[m.encoded] = m.col_name
        return WorkflowDataFrame(*cols)


class WhereVisitor(ExpressionVisitor):
    def __init__(self, context: VisitorContext):
        super().__init__(context)

    def visitWhereClause(self, ctx: qp.WhereClauseContext) -> None:
        self._filter(ctx)

    def visitHavingClause(self, ctx: qp.HavingClauseContext):
        self._filter(ctx)

    def _filter(self, ctx: Any):
        cond = self._get_single_column(ctx.booleanExpression())
        current = self.workflow.op_to_df(
            list(self.current.keys()), "filter_df", self.current, cond
        )
        self.update_current(current)


class AggregationVisitor(VisitorBase):
    def __init__(self, context: VisitorContext):
        super().__init__(context)
        self._internal_exprs: Dict[str, Column] = {}
        self._gp_map: Dict[str, Tuple[str, AggFunctionSpec]] = {}
        self._gp_keys: List[str] = []
        self._f_to_name: Dict[int, str] = {}

    def visitFirst(  # type: ignore
        self, ctx: qp.FirstContext
    ) -> Tuple[AggFunctionSpec, List[Any]]:
        args = [ctx.expression()]
        dropna = ctx.IGNORE() is not None
        func = AggFunctionSpec("first", unique=False, dropna=dropna)
        return func, args

    def visitLast(  # type: ignore
        self, ctx: qp.LastContext
    ) -> Tuple[AggFunctionSpec, List[Any]]:
        args = [ctx.expression()]
        dropna = ctx.IGNORE() is not None
        func = AggFunctionSpec("last", unique=False, dropna=dropna)
        return func, args

    def visitFunctionCall(  # type: ignore
        self, ctx: qp.FunctionCallContext
    ) -> Tuple[AggFunctionSpec, List[Any]]:
        func_name = self.visitFunctionName(ctx.functionName())
        args = ctx.argument
        unique = (
            ctx.setQuantifier() is not None
            and ctx.setQuantifier().DISTINCT() is not None
        )
        func = AggFunctionSpec(
            func_name,
            unique=unique,
            dropna=func_name not in ["first", "last", "first_value", "last_value"],
        )
        return func, args

    def visitRegularQuerySpecification(
        self, ctx: qp.RegularQuerySpecificationContext
    ) -> None:
        self._handle_mentions()
        self._handle_agg_funcs()
        self._handle_group_by()
        if len(self._internal_exprs) > 0:
            self.update_current(
                WorkflowDataFrame(self.current, *list(self._internal_exprs.values()))
            )
        self.update_current(
            self.workflow.op_to_df(
                list(self._gp_map.keys()),
                "group_agg",
                self.current,
                self._gp_keys,
                self._gp_map,
            )
        )
        self.set("agg_func_to_col", self._f_to_name)

    def _handle_mentions(self) -> None:
        for m in self.get("non_agg_mentions"):
            func = AggFunctionSpec("first", unique=False, dropna=False)
            self._gp_map[m.encoded] = (m.encoded, func)

    def _handle_agg_funcs(self) -> None:
        for f_ctx in self.get("agg_funcs", None):
            func, args = self.visit(f_ctx)
            name = self._obj_to_col_name(self.expr(f_ctx))
            if func.name != "count":
                self.assert_support(len(args) == 1, f_ctx)
                internal_expr = self._get_func_args(args)
            else:
                if len(args) == 0:
                    expr = "*"
                else:
                    expr = self.expr(args[0])
                if expr == "*":
                    self.assert_support(len(args) <= 1, f_ctx)
                    internal_expr = ["*"]
                    func = AggFunctionSpec(func.name, func.unique, dropna=False)
                else:
                    internal_expr = self._get_func_args(args)
                    func = AggFunctionSpec(func.name, func.unique, dropna=True)
            self._gp_map[name] = (",".join(internal_expr), func)
            self._f_to_name[id(f_ctx)] = name

    def _handle_group_by(self) -> None:
        for i_ctx in self.get("group_by_expressions"):
            internal_expr, _ = self._get_internal_col(i_ctx)
            self._gp_keys.append(internal_expr)

    def _get_func_args(self, args: Any) -> List[str]:
        e: List[str] = []
        for i in range(len(args)):
            x, _ = self._get_internal_col(args[i])
            e.append(x)
        return e

    def _get_internal_col(self, ctx: Any) -> Tuple[str, Column]:
        internal_expr = self._obj_to_col_name(self.expr(ctx))
        if internal_expr not in self._internal_exprs:
            col = ExpressionVisitor(self)._get_single_column(ctx)
            self._internal_exprs[internal_expr] = col.rename(internal_expr)
        return internal_expr, self._internal_exprs[internal_expr]


class WindowVisitor(VisitorBase):
    def __init__(self, context: VisitorContext):
        super().__init__(context)
        self._internal_exprs: Dict[str, Column] = {}
        self._windows: Dict[str, Tuple[List[str], WindowFrameSpec]] = {}
        self._f_to_name: Dict[int, str] = {}

    def visitSortItem(self, ctx: qp.SortItemContext) -> OrderItemSpec:
        name, _ = self._get_internal_col(ctx.expression())
        asc = ctx.DESC() is None
        if ctx.FIRST() is None and ctx.LAST() is None:
            na_position = "auto"
        else:
            na_position = "first" if ctx.FIRST() is not None else "last"
        return OrderItemSpec(name, asc, na_position)

    def visitFrameBound(self, ctx: qp.FrameBoundContext) -> Tuple[bool, int]:
        unbounded = ctx.UNBOUNDED() is not None
        preceding = ctx.PRECEDING() is not None
        current = ctx.CURRENT() is not None
        if current:
            return False, 0
        if unbounded:
            return True, 0
        n = int(eval(self.to_str(ctx.expression(), "")))
        return False, -n if preceding else n

    def visitWindowFrame(self, ctx: qp.WindowFrameContext) -> WindowFrameSpec:
        if ctx is None:
            return make_windowframe_spec("")
        self.assert_support(ctx.ROWS() is not None and ctx.BETWEEN() is not None, ctx)
        s_unbounded, s_n = self.visitFrameBound(ctx.start)  # type: ignore
        e_unbounded, e_n = self.visitFrameBound(ctx.end)  # type: ignore
        return make_windowframe_spec(
            "rows", None if s_unbounded else s_n, None if e_unbounded else e_n
        )

    def visitWindowDef(self, ctx: qp.WindowDefContext) -> WindowSpec:
        wf = self.visitWindowFrame(ctx.windowFrame())
        partition_keys = [self._get_internal_col(p)[0] for p in ctx.partition]
        sort = OrderBySpec(*[self.visitSortItem(s) for s in ctx.sortItem()])
        return WindowSpec(
            "", partition_keys=partition_keys, order_by=sort, windowframe=wf
        )

    def visitFunctionCall(  # type: ignore
        self, ctx: qp.FunctionCallContext
    ) -> Tuple[WindowFunctionSpec, List[Any]]:
        func_name = self.visitFunctionName(ctx.functionName())
        args = ctx.argument
        unique = (
            ctx.setQuantifier() is not None
            and ctx.setQuantifier().DISTINCT() is not None
        )
        ws = self.visit(ctx.windowSpec())  # it won't be None
        func = WindowFunctionSpec(func_name, unique=unique, dropna=False, window=ws)
        return func, args

    def visitRegularQuerySpecification(
        self, ctx: qp.RegularQuerySpecificationContext
    ) -> None:
        self._handle_window_funcs()
        if len(self._internal_exprs) > 0:
            self.update_current(
                WorkflowDataFrame(self.current, *list(self._internal_exprs.values()))
            )
        for k, v in self._windows.items():
            self.update_current(
                self.workflow.op_to_df(
                    list(self.current.keys()) + [k],
                    "window",
                    self.current,
                    v[1],
                    v[0],
                    k,
                )
            )
        self.set("window_func_to_col", self._f_to_name)

    def _handle_window_funcs(self) -> None:
        for f_ctx in self.get("window_funcs", None):
            func, args = self.visit(f_ctx)
            name = self._obj_to_col_name(self.expr(f_ctx))
            args = self._get_func_args(args)
            self._windows[name] = (args, func)
            self._f_to_name[id(f_ctx)] = name

    def _get_func_args(self, args: Any) -> List[ArgumentSpec]:
        e: List[ArgumentSpec] = []
        for i in range(len(args)):
            is_col, is_single = self.get("func_arg_types")[id(args[i])]
            if is_col:
                x, _ = self._get_internal_col(args[i])
                e.append(ArgumentSpec(True, x))
            else:
                e.append(ArgumentSpec(False, eval(self.to_str(args[i]))))
        return e

    def _get_internal_col(self, ctx: Any) -> Tuple[str, Column]:
        internal_expr = self._obj_to_col_name(self.expr(ctx))
        if internal_expr not in self._internal_exprs:
            col = ExpressionVisitor(self)._get_single_column(ctx)
            self._internal_exprs[internal_expr] = col.rename(internal_expr)
        return internal_expr, self._internal_exprs[internal_expr]


class SelectVisitor(ExpressionVisitor):
    def __init__(self, context: VisitorContext):
        super().__init__(context)

    def visitNamedExpression(self, ctx: qp.NamedExpressionContext) -> Iterable[Column]:
        self.assert_none(ctx.identifierList())
        alias = ""
        if ctx.errorCapturingIdentifier() is not None:
            alias = self.to_str(ctx.errorCapturingIdentifier(), "")
        for col in self.visit(ctx.expression()):
            yield col if alias == "" else col.rename(alias)

    def visitNamedExpressionSeq(
        self, ctx: qp.NamedExpressionSeqContext
    ) -> Iterable[Column]:
        for ne in ctx.namedExpression():
            for r in self.visit(ne):
                yield r

    def visitSelectClause(self, ctx: qp.SelectClauseContext) -> None:
        self.update_current(
            WorkflowDataFrame(*list(self.visit(ctx.namedExpressionSeq())))
        )
        if (
            ctx.setQuantifier() is not None
            and ctx.setQuantifier().DISTINCT() is not None
        ):
            self.update_current(
                self.workflow.op_to_df(
                    list(self.current.keys()), "drop_duplicates", self.current
                )
            )

    def _get_columns(self, ctx: Any) -> List[Column]:
        if self.to_str(ctx) == "*":
            cols: List[Column] = []
            for c in self.current.values():
                cn = self.get("encoded_map", None).get(c.name, "")
                if cn != "":
                    cols.append(c.rename(cn))
            return cols
        return [
            self.current[x.encoded].rename(x.col_name)
            for x in self.get_column_mentions(ctx)
        ]



from typing import Any, Dict, Iterator, List, Set, no_type_check

from triad.utils.assertion import assert_or_throw
from triad.utils.hash import to_uuid

from qpd.constants import AGGREGATION_FUNCTIONS, WINDOW_FUNCTIONS


class SpecBase(object):
    def __init__(self, name: str, **kwargs: Any):
        self._name = name
        self._metadata: Dict[str, Any] = kwargs
        self._uuid = to_uuid(str(type(self)), self._name, self._metadata)

    @property
    def name(self) -> str:
        return self._name

    def __repr__(self) -> str:
        return (self._name, self._metadata).__repr__()

    def __uuid__(self) -> str:
        return self._uuid

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self._uuid == other._uuid

    def __hash__(self) -> int:
        return hash(self._uuid)


class ArgumentSpec(SpecBase):
    def __init__(self, is_col: bool, value: Any):
        super().__init__("", is_col=is_col, value=value)

    @property
    def is_col(self) -> bool:
        return self._metadata["is_col"]

    @property
    def value(self) -> Any:
        return self._metadata["value"]


class OrderItemSpec(SpecBase):
    def __init__(self, name, asc: bool = True, na_position="auto"):
        if na_position == "auto":
            pd_na_position = "first" if asc else "last"
        elif na_position in ["first", "last"]:
            pd_na_position = na_position
        else:
            raise ValueError(f"{na_position} is invalid")
        super().__init__(
            name, asc=asc, na_position=na_position, pd_na_position=pd_na_position
        )

    @property
    def asc(self) -> bool:
        return self._metadata["asc"]

    @property
    def na_position(self) -> str:
        return self._metadata["na_position"]

    @property
    def pd_na_position(self) -> str:
        return self._metadata["pd_na_position"]


class OrderBySpec(SpecBase):
    @no_type_check
    def __init__(self, *items: List[Any]):
        data: List[OrderItemSpec] = []
        names: Set[str] = set()
        first_ct, last_ct = 0, 0
        for i in items:
            if not isinstance(i, OrderItemSpec):
                if isinstance(i, str):
                    i = OrderItemSpec(i)
                else:
                    i = OrderItemSpec(*list(i))
            assert_or_throw(i.name not in names, ValueError(f"{i.name} already exists"))
            names.add(i.name)
            if i.pd_na_position == "first":
                first_ct += 1
            else:
                last_ct += 1
            data.append(i)
        pd_na_position = "last" if last_ct == len(data) else "first"
        super().__init__("", items=data, pd_na_position=pd_na_position)

    @property
    def items(self) -> List[OrderItemSpec]:
        return self._metadata["items"]

    @property
    def keys(self) -> List[str]:
        return [x.name for x in self]

    @property
    def asc(self) -> List[bool]:
        return [x.asc for x in self]  # type: ignore

    @property
    def pd_na_position(self) -> str:
        return self._metadata["pd_na_position"]

    def __iter__(self) -> Iterator[OrderItemSpec]:
        return iter(self.items)

    def __len__(self) -> int:
        return len(self.items)


class PreciateSpec(SpecBase):
    def __init__(self, name: str, positive: bool, **kwargs: Any):
        super().__init__(name, positive=positive, **kwargs)

    @property
    def positive(self) -> bool:
        return self._metadata["positive"]


class IsValueSpec(PreciateSpec):
    def __init__(self, value_expr: str, positive: bool):
        super().__init__(name=value_expr.lower(), positive=positive)

    @property
    def value_expr(self) -> str:
        return self.name


class IsInSpec(PreciateSpec):
    def __init__(self, mode: str, positive: bool, *values: ArgumentSpec):
        mode = mode.lower()
        assert_or_throw(mode in ["in", "between"], ValueError(mode))
        super().__init__(name=mode, positive=positive, values=values)

    @property
    def is_between(self) -> bool:
        return self.name.lower() == "between"

    @property
    def values(self) -> List[ArgumentSpec]:
        return self._metadata["values"]


class FunctionSpec(SpecBase):
    def __init__(self, name: str, unique: bool, dropna: bool, **kwargs: Any):
        super().__init__(name, unique=unique, dropna=dropna, **kwargs)

    @property
    def unique(self) -> bool:
        return self._metadata["unique"]

    @property
    def dropna(self) -> bool:
        return self._metadata["dropna"]


class AggFunctionSpec(FunctionSpec):
    def __init__(self, name: str, unique: bool, dropna: bool):
        assert_or_throw(
            name in AGGREGATION_FUNCTIONS,
            ValueError(f"{name} is not an aggregation function"),
        )
        super().__init__(name, unique=unique, dropna=dropna)


class WindowFrameSpec(SpecBase):
    def __init__(self, frame_type: str, start: Any = None, end: Any = None):
        super().__init__(frame_type, start=start, end=end)

    @property
    def start(self) -> Any:
        return self._metadata["start"]

    @property
    def end(self) -> Any:
        return self._metadata["end"]


class NoWindowFrame(WindowFrameSpec):
    def __init__(self):
        super().__init__("")


class RollingWindowFrame(WindowFrameSpec):
    def __init__(self, start: Any, end: Any):
        super().__init__("rolling", start=start, end=end)
        window = None if start is None or end is None else end - start + 1
        self._metadata["window"] = window
        assert_or_throw(
            self.window is None or self.window > 0, ValueError(f"invalid {start} {end}")
        )

    @property
    def window(self) -> int:
        return self._metadata["window"]


def make_windowframe_spec(
    frame_type: str, start: Any = None, end: Any = None
) -> WindowFrameSpec:
    frame_type = frame_type.lower()
    if frame_type == "":
        return NoWindowFrame()
    if frame_type == "rows":
        if isinstance(start, str):
            start = int(start)
        if isinstance(end, str):
            end = int(end)
        if start is None and end is None:
            return NoWindowFrame()
        return RollingWindowFrame(start, end)
    raise NotImplementedError(frame_type)


class WindowSpec(SpecBase):
    def __init__(
        self,
        name: str,
        partition_keys: List[str],
        order_by: OrderBySpec,
        windowframe: WindowFrameSpec,
    ):
        self._spec_uuid = to_uuid(partition_keys, order_by, windowframe)
        if name == "":
            name = "_" + self.spec_uuid.split("-")[-1]
        super().__init__(
            name,
            partition_keys=partition_keys,
            order_by=order_by,
            windowframe=windowframe,
        )

    @property
    def spec_uuid(self) -> str:
        return self._spec_uuid

    @property
    def partition_keys(self) -> List[str]:
        return self._metadata["partition_keys"]

    @property
    def order_by(self) -> OrderBySpec:
        return self._metadata["order_by"]

    @property
    def windowframe(self) -> WindowFrameSpec:
        return self._metadata["windowframe"]


class WindowFunctionSpec(FunctionSpec):
    def __init__(self, name: str, unique: bool, dropna: bool, window: WindowSpec):
        assert_or_throw(
            name in WINDOW_FUNCTIONS, ValueError(f"{name} is not an window function")
        )
        super().__init__(name, unique=unique, dropna=dropna, window=window)

    @property
    def window(self) -> WindowSpec:
        return self._metadata["window"]

    @property
    def has_windowframe(self) -> bool:
        return not isinstance(self.window.windowframe, NoWindowFrame)

    @property
    def has_partition_by(self) -> bool:
        return len(self.window.partition_keys) > 0

    @property
    def has_order_by(self) -> bool:
        return len(self.window.order_by) > 0



from typing import Any

import pandas as pd

from qpd.dataframe import Column, DataFrame
from qpd.specs import (
    AggFunctionSpec,
    ArgumentSpec,
    IsValueSpec,
    OrderBySpec,
    WindowFunctionSpec,
    WindowSpec,
    make_windowframe_spec,
)
from qpd_test.tests_base import TestsBase
from qpd_test.utils import assert_df_eq
from pytest import raises


class QPDEngineTests(object):
    class Tests(TestsBase):
        def to_df(self, data: Any, cols: Any) -> DataFrame:
            return self.op.to_df(self.to_native_df(data, cols))

        def to_native_df(self, data: Any, columns: Any) -> Any:  # pragma: no cover
            raise NotImplementedError

        def test_parse_join_types(self):
            assert "cross" == self.op.pl_utils.parse_join_type("CROss")
            assert "inner" == self.op.pl_utils.parse_join_type("join")
            assert "inner" == self.op.pl_utils.parse_join_type("Inner")
            assert "left_outer" == self.op.pl_utils.parse_join_type("left")
            assert "left_outer" == self.op.pl_utils.parse_join_type("left  outer")
            assert "right_outer" == self.op.pl_utils.parse_join_type("right")
            assert "right_outer" == self.op.pl_utils.parse_join_type("right_ outer")
            assert "full_outer" == self.op.pl_utils.parse_join_type("full")
            assert "full_outer" == self.op.pl_utils.parse_join_type(" outer ")
            assert "full_outer" == self.op.pl_utils.parse_join_type("full_outer")
            assert "left_anti" == self.op.pl_utils.parse_join_type("anti")
            assert "left_anti" == self.op.pl_utils.parse_join_type("left anti")
            assert "left_semi" == self.op.pl_utils.parse_join_type("semi")
            assert "left_semi" == self.op.pl_utils.parse_join_type("left semi")
            raises(
                NotImplementedError,
                lambda: self.op.pl_utils.parse_join_type("right semi"),
            )

        def test_comparison_op(self):
            def assert_eq(df, first, second, op, expected):
                if isinstance(first, str):
                    first = df[first]
                else:
                    first = Column(first)
                if isinstance(second, str):
                    second = df[second]
                else:
                    second = Column(second)
                self.assert_column_eq(
                    self.op.comparison_op(first, second, op), expected
                )

            assert self.op.comparison_op(Column(0), Column(1), "<").native
            assert not self.op.comparison_op(Column(0), Column(1), ">").native

            df = self.to_df([[1, None], [1, 0], [1, 1], [1, 2]], ["a", "b"])
            assert_eq(df, "a", 0, ">", [True, True, True, True])
            assert_eq(df, "a", "b", ">", [None, True, False, False])
            assert_eq(df, "a", "b", ">=", [None, True, True, False])
            assert_eq(df, "a", "b", "==", [None, False, True, False])
            assert_eq(df, "a", "b", "!=", [None, True, False, True])
            assert_eq(df, "a", "b", "<=", [None, False, True, True])
            assert_eq(df, "a", "b", "<", [None, False, False, True])

            assert_eq(df, 0, "a", ">", [False, False, False, False])
            assert_eq(df, "b", "a", ">", [None, False, False, True])
            assert_eq(df, "b", "a", ">=", [None, False, True, True])
            assert_eq(df, "b", "a", "==", [None, False, True, False])
            assert_eq(df, "b", "a", "!=", [None, True, False, True])
            assert_eq(df, "b", "a", "<=", [None, True, True, False])
            assert_eq(df, "b", "a", "<", [None, True, False, False])

            df = self.to_df([[1, None], [None, 1], [None, None]], ["a", "b"])
            assert_eq(df, "a", "b", ">", [None, None, None])
            assert_eq(df, "a", "b", "==", [None, None, None])
            assert_eq(df, "a", "b", "!=", [None, None, None])

            df = self.to_df([["x", "y"], ["x", None], ["x", "x"]], ["a", "b"])
            assert_eq(df, "a", "b", "<", [True, None, False])
            assert_eq(df, "a", "b", "==", [False, None, True])
            assert_eq(df, "a", "b", "!=", [True, None, False])

        def test_binary_logical_op(self):
            def assert_eq(df, first, second, op, expected):
                if isinstance(first, str):
                    first = df[first]
                else:
                    first = Column(first)
                if isinstance(second, str):
                    second = df[second]
                else:
                    second = Column(second)
                res = self.to_pandas_series(
                    self.op.binary_logical_op(first, second, op)
                ).tolist()
                res = [True if x > 0 else (False if x == 0 else None) for x in res]
                assert_df_eq(pd.DataFrame({"a": res}), pd.DataFrame({"a": expected}))

            df = self.to_df(
                [
                    [True, True],
                    [False, False],
                    [True, False],
                    [True, None],
                    [False, None],
                ],
                ["a", "b"],
            )
            assert_eq(df, "b", True, "and", [True, False, False, None, None])
            assert_eq(df, "b", False, "and", [False, False, False, False, False])
            assert_eq(df, "b", True, "or", [True, True, True, True, True])
            assert_eq(df, "b", False, "or", [True, False, False, None, None])

            assert_eq(df, "b", "a", "and", [True, False, False, None, False])
            assert_eq(df, "b", "a", "or", [True, False, True, True, None])

        def test_is_value(self):
            def assert_eq(values, expr, positive, expected):
                col = self.op.to_col(values)
                res = self.to_pandas_series(
                    self.op.is_value(col, IsValueSpec(expr, positive))
                )
                assert expected == res.tolist()

            assert_eq([1, 2, None], "null", False, [True, True, False])
            assert_eq([1, 2, None], "null", True, [False, False, True])
            assert_eq([True, False, None], "true", True, [True, False, False])
            assert_eq([True, False, None], "true", False, [False, True, True])
            assert_eq([True, False, None], "false", True, [False, True, False])
            assert_eq([True, False, None], "false", False, [True, False, True])

        def test_is_in(self):
            def assert_eq(df, col, positive, collection, expected):
                comp = [df[x] if isinstance(x, str) else x for x in collection]
                self.assert_column_eq(
                    self.op.is_in(df[col], *comp, positive=positive), expected
                )

            df = self.to_df([[0, 1, 1], [2, 3, 4]], ["a", "b", "c"])
            assert_eq(df, "c", True, [], [False, False])
            assert_eq(df, "c", True, [0, 1, 5], [True, False])
            assert_eq(df, "c", True, ["a", "b"], [True, False])
            assert_eq(df, "c", True, [0, "b"], [True, False])
            assert_eq(df, "c", False, [], [True, True])
            assert_eq(df, "c", False, [0, 1, 5], [False, True])
            assert_eq(df, "c", False, ["a", "b"], [False, True])
            assert_eq(df, "c", False, [0, "b"], [False, True])

            df = self.to_df([[0, 1, None]], ["a", "b", "c"])
            assert_eq(df, "c", True, [], [None])
            assert_eq(df, "c", True, [0, 1, 5], [None])
            assert_eq(df, "c", True, ["a", "b"], [None])
            assert_eq(df, "c", True, [0, "b"], [None])
            assert_eq(df, "c", False, [], [None])
            assert_eq(df, "c", False, [0, 1, 5], [None])
            assert_eq(df, "c", False, ["a", "b"], [None])
            assert_eq(df, "c", False, [0, "b"], [None])

        def test_is_between(self):
            def assert_eq(df, col, lower, upper, positive, expected):
                self.assert_column_eq(
                    self.op.is_between(
                        df[col], df[lower], df[upper], positive=positive
                    ),
                    expected,
                )

            df = self.to_df(
                [[0, 1, 1], [2, 3, 4], [None, 1, 1], [1, None, 1], [1, 1, None]],
                ["a", "b", "c"],
            )
            assert_eq(df, "c", "a", "b", True, [True, False, None, None, None])
            assert_eq(df, "c", "a", "b", False, [False, True, None, None, None])

        def test_case_when(self):
            def assert_eq(df, *cols, expected):
                def convert(v: Any) -> Any:
                    return df[v] if isinstance(v, str) else Column(v)

                cols = [convert(x) for x in cols]
                self.assert_column_eq(self.op.case_when(*cols), expected)

            a = self.to_df(
                [
                    [False, False, 1],
                    [False, True, 2],
                    [True, False, 3],
                    [True, True, 4],
                ],
                ["c1", "c2", "v"],
            )
            assert_eq(a, "c1", 10, "v", expected=[1, 2, 10, 10])
            assert_eq(a, "c1", 10, "c2", 11, "v", expected=[1, 11, 10, 10])
            assert_eq(a, "c1", 10, "c2", "v", 11, expected=[11, 2, 10, 10])

        def test_drop_duplicates(self):
            def assert_eq(df, expected, expected_cols):
                res = self.to_pandas_df(self.op.drop_duplicates(df))
                assert_df_eq(res, expected, expected_cols)

            a = self.to_df([["x", "a"], ["x", "a"], [None, None]], ["a", "b"])
            assert_eq(a, [["x", "a"], [None, None]], ["a", "b"])

        def test_order_by_limit(self):
            def assert_eq(df, order_by, limit, expected, expected_cols):
                res = self.to_pandas_df(
                    self.op.order_by_limit(df, OrderBySpec(*order_by), limit)
                )
                assert_df_eq(res, expected, expected_cols)

            a = self.to_df([["x", "a"], ["x", "a"], [None, None]], ["a", "b"])
            assert_eq(a, ["a"], 1, [[None, None]], ["a", "b"])
            assert_eq(a, ["a"], 2, [[None, None], ["x", "a"]], ["a", "b"])
            assert_eq(a, [("a", True, "last")], 2, [["x", "a"], ["x", "a"]], ["a", "b"])
            assert_eq(
                a, [("a", False, "first")], 2, [["x", "a"], [None, None]], ["a", "b"]
            )

        def test_union(self):
            def assert_eq(df1, df2, unique, expected, expected_cols):
                res = self.to_pandas_df(self.op.union(df1, df2, unique=unique))
                assert_df_eq(res, expected, expected_cols)

            a = self.to_df([["x", "a"], ["x", "a"], [None, None]], ["a", "b"])
            b = self.to_df([["xx", "aa"], [None, None], ["a", "x"]], ["b", "a"])
            assert_eq(
                a,
                b,
                False,
                [
                    ["x", "a"],
                    ["x", "a"],
                    [None, None],
                    ["xx", "aa"],
                    [None, None],
                    ["a", "x"],
                ],
                ["a", "b"],
            )
            assert_eq(
                a,
                b,
                True,
                [["x", "a"], ["xx", "aa"], [None, None], ["a", "x"]],
                ["a", "b"],
            )

        def test_intersect(self):
            def assert_eq(df1, df2, unique, expected, expected_cols):
                res = self.to_pandas_df(self.op.intersect(df1, df2, unique=unique))
                assert_df_eq(res, expected, expected_cols)

            a = self.to_df([["x", "a"], ["x", "a"], [None, None]], ["a", "b"])
            b = self.to_df(
                [["xx", "aa"], [None, None], [None, None], ["a", "x"]], ["b", "a"]
            )
            assert_eq(a, b, False, [[None, None]], ["a", "b"])
            assert_eq(a, b, True, [[None, None]], ["a", "b"])
            b = self.to_df([["xx", "aa"], [None, None], ["x", "a"]], ["b", "a"])
            assert_eq(a, b, False, [["x", "a"], ["x", "a"], [None, None]], ["a", "b"])
            assert_eq(a, b, True, [["x", "a"], [None, None]], ["a", "b"])

        def test_except(self):
            def assert_eq(df1, df2, unique, expected, expected_cols):
                res = self.to_pandas_df(self.op.except_df(df1, df2, unique=unique))
                assert_df_eq(res, expected, expected_cols)

            a = self.to_df([["x", "a"], ["x", "a"], [None, None]], ["a", "b"])
            b = self.to_df([["xx", "aa"], [None, None], ["a", "x"]], ["b", "a"])
            assert_eq(a, b, False, [["x", "a"], ["x", "a"]], ["a", "b"])
            assert_eq(a, b, True, [["x", "a"]], ["a", "b"])
            b = self.to_df([["xx", "aa"], [None, None], ["x", "a"]], ["b", "a"])
            assert_eq(a, b, False, [], ["a", "b"])
            assert_eq(a, b, True, [], ["a", "b"])

        def test_joins(self):
            def assert_eq(df1, df2, join_type, on, expected, expected_cols):
                res = self.to_pandas_df(
                    self.op.join(df1, df2, join_type=join_type, on=on)
                )
                assert_df_eq(res, expected, expected_cols)

            df1 = self.to_df([[0, 1], [2, 3]], ["a", "b"])
            df2 = self.to_df([[0, 10], [20, 30]], ["a", "c"])
            df3 = self.to_df([[0, 1], [None, 3]], ["a", "b"])
            df4 = self.to_df([[0, 10], [None, 30]], ["a", "c"])
            assert_eq(df1, df2, "inner", ["a"], [[0, 1, 10]], ["a", "b", "c"])
            assert_eq(df3, df4, "inner", ["a"], [[0, 1, 10]], ["a", "b", "c"])
            assert_eq(df1, df2, "left_semi", ["a"], [[0, 1]], ["a", "b"])
            assert_eq(df3, df4, "left_semi", ["a"], [[0, 1]], ["a", "b"])
            assert_eq(df1, df2, "left_anti", ["a"], [[2, 3]], ["a", "b"])
            assert_eq(df3, df4, "left_anti", ["a"], [[None, 3]], ["a", "b"])
            assert_eq(
                df1,
                df2,
                "left_outer",
                ["a"],
                [[0, 1, 10], [2, 3, None]],
                ["a", "b", "c"],
            )
            assert_eq(
                df3,
                df4,
                "left_outer",
                ["a"],
                [[0, 1, 10], [None, 3, None]],
                ["a", "b", "c"],
            )
            assert_eq(
                df1,
                df2,
                "right_outer",
                ["a"],
                [[0, 1, 10], [20, None, 30]],
                ["a", "b", "c"],
            )
            assert_eq(
                df3,
                df4,
                "right_outer",
                ["a"],
                [[0, 1, 10], [None, None, 30]],
                ["a", "b", "c"],
            )
            assert_eq(
                df1,
                df2,
                "full_outer",
                ["a"],
                [[0, 1, 10], [2, 3, None], [20, None, 30]],
                ["a", "b", "c"],
            )
            assert_eq(
                df3,
                df4,
                "full_outer",
                ["a"],
                [[0, 1, 10], [None, 3, None], [None, None, 30]],
                ["a", "b", "c"],
            )

            df1 = self.to_df([[0, 1], [None, 3]], ["a", "b"])
            df2 = self.to_df([[0, 10], [None, 30]], ["c", "d"])
            assert_eq(
                df1,
                df2,
                "cross",
                [],
                [
                    [0, 1, 0, 10],
                    [None, 3, 0, 10],
                    [0, 1, None, 30],
                    [None, 3, None, 30],
                ],
                ["a", "b", "c", "d"],
            )

        def test_sort(self):
            def ob(*order_by):
                ws = WindowSpec(
                    "",
                    partition_keys=[],
                    order_by=OrderBySpec(*order_by),
                    windowframe=make_windowframe_spec(""),
                )
                return WindowFunctionSpec("row_number", False, False, ws)

            def assert_eq(df, index, *order_by):
                ndf = self.op.to_native(df)
                e = ndf[ndf.columns]
                e["x"] = self.op.to_col(index).native
                self.w_eq(
                    df, ob(*order_by), [], "x", e.values.tolist(), list(e.columns)
                )

            df = self.to_df(
                [[0, 10, None], [1, None, 100], [None, 20, 300]], list("abc")
            )
            assert_eq(df, [1, 2, 3])
            assert_eq(df, [2, 3, 1], ("a", True))
            assert_eq(df, [2, 1, 3], ("a", False))
            assert_eq(df, [2, 3, 1], ("a", True, "first"))
            assert_eq(df, [1, 2, 3], ("a", True, "last"))
            assert_eq(df, [3, 2, 1], ("a", False, "first"))
            assert_eq(df, [2, 1, 3], ("a", False, "last"))

            df = self.to_df(
                [[0, 10, None], [0, 5, None], [0, None, 100], [None, 20, 300]],
                list("abc"),
            )
            assert_eq(df, [4, 3, 2, 1], ("a", True), ("b", True))
            assert_eq(df, [2, 3, 4, 1], ("a", True), ("b", False))
            assert_eq(df, [3, 2, 4, 1], ("a", True), ("b", True, "last"))
            assert_eq(df, [3, 4, 2, 1], ("a", True), ("b", False, "first"))
            assert_eq(df, [2, 3, 1, 4], ("a", True, "last"), ("b", False, "first"))

        def test_agg_first_last(self):
            def assert_eq(df, gp_keys, gp_map, expected, expected_cols):
                res = self.to_pandas_df(self.op.group_agg(df, gp_keys, gp_map))
                assert_df_eq(res, expected, expected_cols)

            pdf = self.to_df([[0, 1], [0, 1], [1, None], [1, 5], [0, None]], ["a", "b"])
            assert_eq(
                pdf,
                ["a"],
                {
                    "b1": ("b", AggFunctionSpec("first", unique=False, dropna=False)),
                    "b2": ("b", AggFunctionSpec("first", unique=False, dropna=True)),
                    "b3": ("b", AggFunctionSpec("last", unique=False, dropna=False)),
                    "b4": ("b", AggFunctionSpec("last", unique=False, dropna=True)),
                },
                [[1.0, 1.0, None, 1.0], [None, 5.0, 5.0, 5.0]],
                ["b1", "b2", "b3", "b4"],
            )

        def test_agg_sum_mean(self):
            def assert_eq(df, gp_keys, gp_map, expected, expected_cols):
                res = self.to_pandas_df(self.op.group_agg(df, gp_keys, gp_map))
                assert_df_eq(res, expected, expected_cols)

            pdf = self.to_df([[0, 1], [0, 1], [0, 2], [1, 5], [0, None]], ["a", "b"])
            assert_eq(
                pdf,
                ["a"],
                {
                    "b1": ("b", AggFunctionSpec("sum", unique=False, dropna=False)),
                    "b2": ("b", AggFunctionSpec("sum", unique=True, dropna=False)),
                    "b3": ("b", AggFunctionSpec("sum", unique=True, dropna=True)),
                    "b4": ("b", AggFunctionSpec("sum", unique=False, dropna=True)),
                },
                [[4, 3, 3, 4], [5, 5, 5, 5]],
                ["b1", "b2", "b3", "b4"],
            )
            pdf = self.to_df([[0, 1], [0, 1], [0, 4], [1, 5], [0, None]], ["a", "b"])
            assert_eq(
                pdf,
                ["a"],
                {
                    "b1": ("b", AggFunctionSpec("avg", unique=False, dropna=False)),
                    "b2": ("b", AggFunctionSpec("avg", unique=True, dropna=False)),
                    "b3": ("b", AggFunctionSpec("mean", unique=True, dropna=True)),
                    "b4": ("b", AggFunctionSpec("mean", unique=False, dropna=True)),
                },
                [[2, 2.5, 2.5, 2], [5, 5, 5, 5]],
                ["b1", "b2", "b3", "b4"],
            )
            assert_eq(
                pdf,
                ["a"],
                {
                    "b1": ("b", AggFunctionSpec("sum", unique=True, dropna=False)),
                    "b2": ("b", AggFunctionSpec("avg", unique=True, dropna=False)),
                },
                [[5, 2.5], [5, 5]],
                ["b1", "b2"],
            )

        def test_agg_min_max(self):
            def assert_eq(df, gp_keys, gp_map, expected, expected_cols):
                res = self.to_pandas_df(self.op.group_agg(df, gp_keys, gp_map))
                assert_df_eq(res, expected, expected_cols)

            pdf = self.to_df([[0, 1], [0, 1], [0, 2], [1, None], [0, None]], ["a", "b"])
            assert_eq(
                pdf,
                ["a"],
                {
                    "b1": ("b", AggFunctionSpec("min", unique=False, dropna=False)),
                    "b2": ("b", AggFunctionSpec("min", unique=True, dropna=False)),
                    "b3": ("b", AggFunctionSpec("min", unique=True, dropna=True)),
                    "b4": ("b", AggFunctionSpec("min", unique=False, dropna=True)),
                },
                [[1.0, 1.0, 1.0, 1.0], [None, None, None, None]],
                ["b1", "b2", "b3", "b4"],
            )
            assert_eq(
                pdf,
                ["a"],
                {
                    "b1": ("b", AggFunctionSpec("max", unique=False, dropna=False)),
                    "b2": ("b", AggFunctionSpec("max", unique=True, dropna=False)),
                    "b3": ("b", AggFunctionSpec("max", unique=True, dropna=True)),
                    "b4": ("b", AggFunctionSpec("max", unique=False, dropna=True)),
                },
                [[2.0, 2.0, 2.0, 2.0], [None, None, None, None]],
                ["b1", "b2", "b3", "b4"],
            )

            pdf = self.to_df(
                [[0, "a"], [0, "a"], [0, "b"], [1, None], [0, None]], ["a", "b"]
            )
            assert_eq(
                pdf,
                ["a"],
                {
                    "b1": ("b", AggFunctionSpec("min", unique=False, dropna=False)),
                    "b2": ("b", AggFunctionSpec("min", unique=True, dropna=False)),
                    "b3": ("b", AggFunctionSpec("min", unique=True, dropna=True)),
                    "b4": ("b", AggFunctionSpec("min", unique=False, dropna=True)),
                },
                [["a", "a", "a", "a"], [None, None, None, None]],
                ["b1", "b2", "b3", "b4"],
            )
            assert_eq(
                pdf,
                ["a"],
                {
                    "b1": ("b", AggFunctionSpec("max", unique=False, dropna=False)),
                    "b2": ("b", AggFunctionSpec("max", unique=True, dropna=False)),
                    "b3": ("b", AggFunctionSpec("max", unique=True, dropna=True)),
                    "b4": ("b", AggFunctionSpec("max", unique=False, dropna=True)),
                },
                [["b", "b", "b", "b"], [None, None, None, None]],
                ["b1", "b2", "b3", "b4"],
            )

        def test_agg_count(self):
            def assert_eq(df, gp_keys, gp_map, expected, expected_cols):
                res = self.to_pandas_df(self.op.group_agg(df, gp_keys, gp_map))
                assert_df_eq(res, expected, expected_cols)

            pdf = self.to_df([[0, 1], [0, 1], [0, 4], [1, None], [0, None]], ["a", "b"])
            # count on single column
            assert_eq(
                pdf,
                ["a"],
                {
                    "b1": ("b", AggFunctionSpec("count", unique=False, dropna=False)),
                    "b2": ("b", AggFunctionSpec("count", unique=True, dropna=False)),
                    "b3": ("b", AggFunctionSpec("count", unique=False, dropna=True)),
                    "b4": ("b", AggFunctionSpec("count", unique=True, dropna=True)),
                },
                [[4, 3, 3, 2], [1, 1, 0, 0]],
                ["b1", "b2", "b3", "b4"],
            )
            # count on all columns with nulls
            assert_eq(
                pdf,
                ["a"],
                {
                    "b1": ("*", AggFunctionSpec("count", unique=False, dropna=False)),
                    "b2": ("*", AggFunctionSpec("count", unique=True, dropna=False)),
                },
                [[4, 3], [1, 1]],
                ["b1", "b2"],
            )
            # count on keys with nulls
            assert_eq(
                pdf,
                ["b"],
                {
                    "a1": ("*", AggFunctionSpec("count", unique=False, dropna=False)),
                    "a2": ("*", AggFunctionSpec("count", unique=True, dropna=False)),
                },
                [[2, 2], [2, 1], [1, 1]],
                ["a1", "a2"],
            )
            # count multiple cols
            pdf = self.to_df(
                [
                    [0, 1, None],
                    [0, 1, None],
                    [0, 4, 1],
                    [1, None, None],
                    [None, None, None],
                ],
                ["a", "b", "c"],
            )
            assert_eq(
                pdf,
                ["a"],
                {
                    "a1": ("b,c", AggFunctionSpec("count", unique=False, dropna=False)),
                    "a2": ("b,c", AggFunctionSpec("count", unique=True, dropna=False)),
                },
                [[3, 2], [1, 1], [1, 1]],
                ["a1", "a2"],
            )

        def test_agg_combined(self):
            def assert_eq(df, gp_keys, gp_map, expected, expected_cols):
                res = self.to_pandas_df(self.op.group_agg(df, gp_keys, gp_map))
                assert_df_eq(res, expected, expected_cols)

            # combined case
            pdf = self.to_df(
                [
                    [0, 1, None],
                    [0, 1, None],
                    [0, 4, 1],
                    [1, None, None],
                    [None, None, None],
                ],
                ["a", "b", "c"],
            )
            assert_eq(
                pdf,
                ["a"],
                {
                    "a1": ("a", AggFunctionSpec("sum", unique=False, dropna=False)),
                    "a2": ("*", AggFunctionSpec("count", unique=True, dropna=False)),
                    "a3": ("c", AggFunctionSpec("sum", unique=True, dropna=False)),
                    "a4": ("*", AggFunctionSpec("count", unique=False, dropna=False)),
                    "a5": ("c", AggFunctionSpec("min", unique=False, dropna=False)),
                    "a6": ("b,c", AggFunctionSpec("count", unique=True, dropna=False)),
                },
                [
                    [0.0, 2.0, 1, 3.0, 1.0, 2.0],
                    [1.0, 1.0, None, 1.0, None, 1.0],
                    [None, 1.0, None, 1.0, None, 1.0],
                ],
                ["a1", "a2", "a3", "a4", "a5", "a6"],
            )

        def test_agg_no_group_first_last(self):
            def assert_eq(df, gp_map, expected, expected_cols):
                res = self.to_pandas_df(self.op.group_agg(df, [], gp_map))
                assert_df_eq(res, expected, expected_cols)

            pdf = self.to_df([[0, 1], [0, 1], [1, None], [1, 5], [0, None]], ["a", "b"])
            assert_eq(
                pdf,
                {
                    "b1": ("b", AggFunctionSpec("first", unique=False, dropna=False)),
                    "b2": ("b", AggFunctionSpec("first", unique=False, dropna=True)),
                    "b3": ("b", AggFunctionSpec("last", unique=False, dropna=False)),
                    "b4": ("b", AggFunctionSpec("last", unique=False, dropna=True)),
                },
                [[1.0, 1.0, None, 5.0]],
                ["b1", "b2", "b3", "b4"],
            )

        def test_agg_no_group_sum(self):
            def assert_eq(df, gp_map, expected, expected_cols):
                res = self.to_pandas_df(self.op.group_agg(df, [], gp_map))
                assert_df_eq(res, expected, expected_cols)

            pdf = self.to_df([[0, 1], [0, 1], [1, None], [1, 5], [0, None]], ["a", "b"])
            assert_eq(
                pdf,
                {
                    "b1": ("b", AggFunctionSpec("sum", unique=False, dropna=False)),
                    "b2": ("b", AggFunctionSpec("sum", unique=True, dropna=False)),
                    "b3": ("b", AggFunctionSpec("sum", unique=True, dropna=True)),
                    "b4": ("b", AggFunctionSpec("sum", unique=False, dropna=True)),
                },
                [[7, 6, 6, 7]],
                ["b1", "b2", "b3", "b4"],
            )

        def test_agg_no_group_min_max(self):
            def assert_eq(df, gp_map, expected, expected_cols):
                res = self.to_pandas_df(self.op.group_agg(df, [], gp_map))
                assert_df_eq(res, expected, expected_cols)

            pdf = self.to_df([[0, 1], [0, 1], [1, None], [1, 5], [0, None]], ["a", "b"])
            assert_eq(
                pdf,
                {
                    "b1": ("b", AggFunctionSpec("min", unique=False, dropna=False)),
                    "b2": ("b", AggFunctionSpec("min", unique=True, dropna=False)),
                    "b3": ("b", AggFunctionSpec("min", unique=True, dropna=True)),
                    "b4": ("b", AggFunctionSpec("min", unique=False, dropna=True)),
                },
                [[1.0, 1.0, 1.0, 1.0]],
                ["b1", "b2", "b3", "b4"],
            )
            assert_eq(
                pdf,
                {
                    "b1": ("b", AggFunctionSpec("max", unique=False, dropna=False)),
                    "b2": ("b", AggFunctionSpec("max", unique=True, dropna=False)),
                    "b3": ("b", AggFunctionSpec("max", unique=True, dropna=True)),
                    "b4": ("b", AggFunctionSpec("max", unique=False, dropna=True)),
                },
                [[5, 5, 5, 5]],
                ["b1", "b2", "b3", "b4"],
            )

        def test_agg_no_group_mean(self):
            def assert_eq(df, gp_map, expected, expected_cols):
                res = self.to_pandas_df(self.op.group_agg(df, [], gp_map))
                assert_df_eq(res, expected, expected_cols)

            pdf = self.to_df([[0, 1], [0, 1], [0, 4], [0, None]], ["a", "b"])
            assert_eq(
                pdf,
                {
                    "b1": ("b", AggFunctionSpec("avg", unique=False, dropna=False)),
                    "b2": ("b", AggFunctionSpec("avg", unique=True, dropna=False)),
                    "b3": ("b", AggFunctionSpec("mean", unique=True, dropna=True)),
                    "b4": ("b", AggFunctionSpec("mean", unique=False, dropna=True)),
                },
                [[2, 2.5, 2.5, 2]],
                ["b1", "b2", "b3", "b4"],
            )

        def test_agg_no_group_count(self):
            def assert_eq(df, gp_map, expected, expected_cols):
                res = self.to_pandas_df(self.op.group_agg(df, [], gp_map))
                assert_df_eq(res, expected, expected_cols)

            pdf = self.to_df([[0, 1], [0, 1], [1, None]], ["a", "b"])
            # count on single column
            assert_eq(
                pdf,
                {
                    "b1": ("b", AggFunctionSpec("count", unique=False, dropna=False)),
                    "b2": ("b", AggFunctionSpec("count", unique=True, dropna=False)),
                    "b3": ("b", AggFunctionSpec("count", unique=False, dropna=True)),
                    "b4": ("b", AggFunctionSpec("count", unique=True, dropna=True)),
                },
                [[3, 2, 2, 1]],
                ["b1", "b2", "b3", "b4"],
            )
            # count on all columns with nulls
            assert_eq(
                pdf,
                {
                    "b1": ("*", AggFunctionSpec("count", unique=False, dropna=False)),
                    "b2": ("*", AggFunctionSpec("count", unique=True, dropna=False)),
                },
                [[3, 2]],
                ["b1", "b2"],
            )
            # count multiple cols
            pdf = self.to_df(
                [
                    [0, 1, None],
                    [0, 1, None],
                    [0, 4, 1],
                    [1, None, None],
                    [None, None, None],
                ],
                ["a", "b", "c"],
            )
            assert_eq(
                pdf,
                {
                    "a1": ("b,c", AggFunctionSpec("count", unique=False, dropna=False)),
                    "a2": ("b,c", AggFunctionSpec("count", unique=True, dropna=False)),
                },
                [[5, 3]],
                ["a1", "a2"],
            )

        def test_agg_no_group_combined(self):
            def assert_eq(df, gp_map, expected, expected_cols):
                res = self.to_pandas_df(self.op.group_agg(df, [], gp_map))
                assert_df_eq(res, expected, expected_cols)

            pdf = self.to_df(
                [
                    [0, 1, None],
                    [0, 1, None],
                    [0, 4, 1],
                    [1, None, None],
                    [None, None, None],
                ],
                ["a", "b", "c"],
            )
            # combo test
            assert_eq(
                pdf,
                {
                    "a1": ("b,c", AggFunctionSpec("count", unique=False, dropna=False)),
                    "a2": ("*", AggFunctionSpec("count", unique=True, dropna=False)),
                    "a3": ("a", AggFunctionSpec("first", unique=True, dropna=False)),
                    "a4": ("b", AggFunctionSpec("max", unique=True, dropna=False)),
                },
                [[5, 4, 0, 4]],
                ["a1", "a2", "a3", "a4"],
            )

        def test_window_row_number(self):
            def make_func(partition_keys, order_by):
                ws = WindowSpec(
                    "",
                    partition_keys=partition_keys,
                    order_by=OrderBySpec(*order_by),
                    windowframe=make_windowframe_spec(""),
                )
                return WindowFunctionSpec("row_number", False, False, ws)

            a = self.to_df([[0, 1], [0, 2]], ["a", "b"])
            self.w_eq(
                a, make_func([], []), [], "x", [[0, 1, 1], [0, 2, 2]], ["a", "b", "x"]
            )
            self.w_eq(
                a,
                make_func([], [("a", True), ("b", False)]),
                [],
                "x",
                [[0, 1, 2], [0, 2, 1]],
                ["a", "b", "x"],
            )
            a = self.to_df([[0, 1], [0, None], [None, 1], [None, None]], ["a", "b"])
            self.w_eq(
                a,
                make_func([], [("a", True), ("b", True)]),
                [],
                "x",
                [[0, 1, 4], [0, None, 3], [None, 1, 2], [None, None, 1]],
                ["a", "b", "x"],
            )
            self.w_eq(
                a,
                make_func([], [("a", True, "last"), ("b", True, "last")]),
                [],
                "x",
                [[0, 1, 1], [0, None, 2], [None, 1, 3], [None, None, 4]],
                ["a", "b", "x"],
            )
            a = self.to_df([[0, 1], [0, 4], [2, 3]], ["a", "b"])
            self.w_eq(
                a,
                make_func([], [("b", False)]),
                [],
                "x",
                [[0, 1, 3], [0, 4, 1], [2, 3, 2]],
                ["a", "b", "x"],
            )

        def test_window_row_number_partition_by(self):
            def make_func(partition_keys, order_by):
                ws = WindowSpec(
                    "",
                    partition_keys=partition_keys,
                    order_by=OrderBySpec(*order_by),
                    windowframe=make_windowframe_spec(""),
                )
                return WindowFunctionSpec("row_number", False, False, ws)

            a = self.to_df([[0, 1], [0, None], [None, 1], [None, None]], ["a", "b"])
            self.w_eq(
                a,
                make_func(["a"], [("a", True), ("b", True)]),
                [],
                "x",
                [[0, 1, 2], [0, None, 1], [None, 1, 2], [None, None, 1]],
                ["a", "b", "x"],
            )

        def test_window_ranks(self):
            def make_func(func, partition_keys, order_by):
                ws = WindowSpec(
                    "",
                    partition_keys=partition_keys,
                    order_by=OrderBySpec(*order_by),
                    windowframe=make_windowframe_spec(""),
                )
                return WindowFunctionSpec(func, False, False, ws)

            a = self.to_df([[0, 1], [0, 1], [None, 1], [0, None]], ["a", "b"])
            self.w_eq(
                a,
                make_func("rank", [], ["b"]),
                [],
                "x",
                [[0, 1, 2], [0, 1, 2], [None, 1, 2], [0, None, 1]],
                ["a", "b", "x"],
            )

            self.w_eq(
                a,
                make_func("rank", [], [("b", False)]),
                [],
                "x",
                [[0, 1, 1], [0, 1, 1], [None, 1, 1], [0, None, 4]],
                ["a", "b", "x"],
            )

            self.w_eq(
                a,
                make_func("dense_rank", [], ["b"]),
                [],
                "x",
                [[0, 1, 2], [0, 1, 2], [None, 1, 2], [0, None, 1]],
                ["a", "b", "x"],
            )

            self.w_eq(
                a,
                make_func("dense_rank", [], [("b", False)]),
                [],
                "x",
                [[0, 1, 1], [0, 1, 1], [None, 1, 1], [0, None, 2]],
                ["a", "b", "x"],
            )

            self.w_eq(
                a,
                make_func("percentile_rank", [], ["b"]),
                [],
                "x",
                [
                    [0, 1, 0.33333333333],
                    [0, 1, 0.33333333333],
                    [None, 1, 0.33333333333],
                    [0, None, 0.0],
                ],
                ["a", "b", "x"],
            )

            self.w_eq(
                a,
                make_func("percentile_rank", [], [("b", False)]),
                [],
                "x",
                [[0, 1, 0.0], [0, 1, 0.0], [None, 1, 0.0], [0, None, 1.0]],
                ["a", "b", "x"],
            )

        def test_window_ranks_partition_by(self):
            def make_func(func, partition_keys, order_by):
                ws = WindowSpec(
                    "",
                    partition_keys=partition_keys,
                    order_by=OrderBySpec(*order_by),
                    windowframe=make_windowframe_spec(""),
                )
                return WindowFunctionSpec(func, False, False, ws)

            a = self.to_df([[0, 1], [0, 1], [None, 1], [0, None]], ["a", "b"])
            self.w_eq(
                a,
                make_func("rank", ["a"], ["b"]),
                [],
                "x",
                [[0, 1, 2], [0, 1, 2], [None, 1, 1], [0, None, 1]],
                ["a", "b", "x"],
            )

            self.w_eq(
                a,
                make_func("rank", ["a"], [("b", False)]),
                [],
                "x",
                [[0, 1, 1], [0, 1, 1], [None, 1, 1], [0, None, 3]],
                ["a", "b", "x"],
            )

            self.w_eq(
                a,
                make_func("dense_rank", ["a"], ["b"]),
                [],
                "x",
                [[0, 1, 2], [0, 1, 2], [None, 1, 1], [0, None, 1]],
                ["a", "b", "x"],
            )

            self.w_eq(
                a,
                make_func("dense_rank", ["a"], [("b", False)]),
                [],
                "x",
                [[0, 1, 1], [0, 1, 1], [None, 1, 1], [0, None, 2]],
                ["a", "b", "x"],
            )

            self.w_eq(
                a,
                make_func("percentile_rank", ["a"], ["b"]),
                [],
                "x",
                [[0, 1, 0.5], [0, 1, 0.5], [None, 1, 0.0], [0, None, 0.0]],
                ["a", "b", "x"],
            )

            self.w_eq(
                a,
                make_func("percentile_rank", ["a"], [("b", False)]),
                [],
                "x",
                [[0, 1, 0.0], [0, 1, 0.0], [None, 1, 0.0], [0, None, 1.0]],
                ["a", "b", "x"],
            )

        def test_window_lead_lag(self):
            def make_func(func, partition_keys, order_by):
                ws = WindowSpec(
                    "",
                    partition_keys=partition_keys,
                    order_by=OrderBySpec(*order_by),
                    windowframe=make_windowframe_spec(""),
                )
                return WindowFunctionSpec(func, False, False, ws)

            a = self.to_df([[0, 1], [0, 1], [None, 1], [0, None]], ["a", "b"])
            self.w_eq(
                a,
                make_func("lag", [], []),
                [col("b"), 1],
                "x",
                [[0, 1, None], [0, 1, 1], [None, 1, 1], [0, None, 1]],
                ["a", "b", "x"],
            )
            self.w_eq(
                a,
                make_func("lag", [], []),
                [col("b"), 1, 100],
                "x",
                [[0, 1, 100], [0, 1, 1], [None, 1, 1], [0, None, 1]],
                ["a", "b", "x"],
            )

            self.w_eq(
                a,
                make_func("lead", [], []),
                [col("b"), 1],
                "x",
                [[0, 1, 1], [0, 1, 1], [None, 1, None], [0, None, None]],
                ["a", "b", "x"],
            )
            self.w_eq(
                a,
                make_func("lead", [], []),
                [col("b"), 1, 100],
                "x",
                [[0, 1, 1], [0, 1, 1], [None, 1, None], [0, None, 100]],
                ["a", "b", "x"],
            )

        def test_window_lead_lag_partition_by(self):
            def make_func(func, partition_keys, order_by):
                ws = WindowSpec(
                    "",
                    partition_keys=partition_keys,
                    order_by=OrderBySpec(*order_by),
                    windowframe=make_windowframe_spec(""),
                )
                return WindowFunctionSpec(func, False, False, ws)

            a = self.to_df([[0, 1], [0, 1], [None, 1], [0, None]], ["a", "b"])
            self.w_eq(
                a,
                make_func("lag", ["a"], ["b"]),
                [col("b"), 1],
                "x",
                [[0, 1, None], [0, 1, 1], [None, 1, None], [0, None, None]],
                ["a", "b", "x"],
            )

            self.w_eq(
                a,
                make_func("lead", ["a"], ["b"]),
                [col("b"), 1],
                "x",
                [[0, 1, 1], [0, 1, None], [None, 1, None], [0, None, 1]],
                ["a", "b", "x"],
            )

        def test_window_sum(self):
            def make_func(partition_keys, order_by, start=None, end=None):
                wtype = "" if start is None and end is None else "rows"
                ws = WindowSpec(
                    "",
                    partition_keys=partition_keys,
                    order_by=OrderBySpec(*order_by),
                    windowframe=make_windowframe_spec(wtype, start, end),
                )
                return WindowFunctionSpec("sum", False, False, ws)

            a = self.to_df([[0, 1], [0, 2]], ["a", "b"])
            # no partition no window frame
            self.w_eq(
                a,
                make_func([], []),
                [col("b")],
                "x",
                [[0, 1, 3], [0, 2, 3]],
                ["a", "b", "x"],
            )
            self.w_eq(
                a,
                make_func([], [("a", True), ("b", False)]),
                [col("b")],
                "x",
                [[0, 1, 3], [0, 2, 3]],
                ["a", "b", "x"],
            )

            a = self.to_df([[0, 1], [None, 1], [0, None], [0, 2]], ["a", "b"])
            # expanding, no partition
            self.w_eq(
                a,
                make_func([], ["b", "a"], end=0),
                [col("b")],
                "x",
                [[0, 1, 2], [None, 1, 1], [0, None, None], [0, 2, 4]],
                ["a", "b", "x"],
            )

            # rolling, no partition
            self.w_eq(
                a,
                make_func([], ["b", "a"], start=-1, end=0),
                [col("b")],
                "x",
                [[0, 1, 2], [None, 1, 1], [0, None, None], [0, 2, 3]],
                ["a", "b", "x"],
            )

            if not pd.__version__.startswith("1.1"):
                # irregular, no partition
                self.w_eq(
                    a,
                    make_func([], ["b", "a"], end=-1),
                    [col("b")],
                    "x",
                    [[0, 1, 1], [None, 1, None], [0, None, None], [0, 2, 2]],
                    ["a", "b", "x"],
                )

                # irregular, no partition
                self.w_eq(
                    a,
                    make_func([], ["b", "a"], start=-2, end=-1),
                    [col("b")],
                    "x",
                    [[0, 1, 1], [None, 1, None], [0, None, None], [0, 2, 2]],
                    ["a", "b", "x"],
                )

                # irregular, no partition
                self.w_eq(
                    a,
                    make_func([], ["b", "a"], start=1, end=None),
                    [col("b")],
                    "x",
                    [[0, 1, 2], [None, 1, 3], [0, None, 4], [0, 2, None]],
                    ["a", "b", "x"],
                )

        def test_window_sum_partition_by(self):
            def make_func(partition_keys, order_by, start=None, end=None):
                wtype = "" if start is None and end is None else "rows"
                ws = WindowSpec(
                    "",
                    partition_keys=partition_keys,
                    order_by=OrderBySpec(*order_by),
                    windowframe=make_windowframe_spec(wtype, start, end),
                )
                return WindowFunctionSpec("sum", False, False, ws)

            # no partition no window frame
            # partition no window frame
            a = self.to_df([[0, 1], [None, 1], [0, None], [0, 2]], ["a", "b"])
            self.w_eq(
                a,
                make_func(["a"], []),
                [col("b")],
                "x",
                [[0, 1, 3], [None, 1, 1], [0, None, 3], [0, 2, 3]],
                ["a", "b", "x"],
            )

            # expanding, with partition
            self.w_eq(
                a,
                make_func(["a"], ["b", "a"], end=0),
                [col("b")],
                "x",
                [[0, 1, 1], [None, 1, 1], [0, None, None], [0, 2, 3]],
                ["a", "b", "x"],
            )

            # rolling, with partition
            self.w_eq(
                a,
                make_func(["a"], ["b", "a"], start=-1, end=0),
                [col("b")],
                "x",
                [[0, 1, 1], [None, 1, 1], [0, None, None], [0, 2, 3]],
                ["a", "b", "x"],
            )
            self.w_eq(
                a,
                make_func(["a"], ["b", "a"], start=0, end=0),
                [col("b")],
                "x",
                [[0, 1, 1], [None, 1, 1], [0, None, None], [0, 2, 2]],
                ["a", "b", "x"],
            )

            if not pd.__version__.startswith("1.1"):
                # irregular, with partition
                self.w_eq(
                    a,
                    make_func(["a"], ["b"], end=-1),
                    [col("b")],
                    "x",
                    [[0, 1, None], [None, 1, None], [0, None, None], [0, 2, 1]],
                    ["a", "b", "x"],
                )

                # irregular, with partition
                self.w_eq(
                    a,
                    make_func(["a"], ["b"], start=-1, end=-1),
                    [col("b")],
                    "x",
                    [[0, 1, None], [None, 1, None], [0, None, None], [0, 2, 1]],
                    ["a", "b", "x"],
                )

                # irregular, with partition
                self.w_eq(
                    a,
                    make_func(["a"], ["b"], start=1, end=None),
                    [col("b")],
                    "x",
                    [[0, 1, 2], [None, 1, None], [0, None, 3], [0, 2, None]],
                    ["a", "b", "x"],
                )

        def test_window_simple_agg_partition_by(self):
            def make_func(func, partition_keys, order_by, start=None, end=None):
                wtype = "" if start is None and end is None else "rows"
                ws = WindowSpec(
                    "",
                    partition_keys=partition_keys,
                    order_by=OrderBySpec(*order_by),
                    windowframe=make_windowframe_spec(wtype, start, end),
                )
                return WindowFunctionSpec(func, False, False, ws)

            a = self.to_df([[0, 1], [None, 1], [0, None], [0, 2]], ["a", "b"])
            self.w_eq(
                a,
                make_func("max", ["a"], ["b"]),
                [col("b")],
                "x",
                [[0, 1, 2], [None, 1, 1], [0, None, 2], [0, 2, 2]],
                ["a", "b", "x"],
            )
            self.w_eq(
                a,
                make_func("max", ["a"], ["b"], start=-1, end=0),
                [col("b")],
                "x",
                [[0, 1, 1], [None, 1, 1], [0, None, None], [0, 2, 2]],
                ["a", "b", "x"],
            )

            self.w_eq(
                a,
                make_func("min", ["a"], ["b"]),
                [col("b")],
                "x",
                [[0, 1, 1], [None, 1, 1], [0, None, 1], [0, 2, 1]],
                ["a", "b", "x"],
            )
            self.w_eq(
                a,
                make_func("min", ["a"], ["b"], start=-1, end=0),
                [col("b")],
                "x",
                [[0, 1, 1], [None, 1, 1], [0, None, None], [0, 2, 1]],
                ["a", "b", "x"],
            )

            self.w_eq(
                a,
                make_func("avg", ["a"], ["b"]),
                [col("b")],
                "x",
                [[0, 1, 1.5], [None, 1, 1], [0, None, 1.5], [0, 2, 1.5]],
                ["a", "b", "x"],
            )
            self.w_eq(
                a,
                make_func("mean", ["a"], ["b"], start=-1, end=0),
                [col("b")],
                "x",
                [[0, 1, 1.0], [None, 1, 1], [0, None, None], [0, 2, 1.5]],
                ["a", "b", "x"],
            )

            self.w_eq(
                a,
                make_func("count", ["a"], ["b"]),
                [col("b")],
                "x",
                [[0, 1, 2], [None, 1, 1], [0, None, 2], [0, 2, 2]],
                ["a", "b", "x"],
            )
            a = self.to_df([[0, 1], [None, 1], [0, None], [0, 3]], ["a", "b"])
            self.w_eq(
                a,
                make_func("count", ["a"], [("b", True)], start=-1, end=0),
                [col("b")],
                "x",
                [[0, 1, 1], [None, 1, 1], [0, None, 0], [0, 3, 2]],
                ["a", "b", "x"],
            )

        def w_eq(self, df, func, cols, dest, expected, expected_cols):
            cols = [
                x if isinstance(x, ArgumentSpec) else ArgumentSpec(False, x)
                for x in cols
            ]
            res = self.to_pandas_df(self.op.window(df, func, cols, dest))
            assert_df_eq(res, expected, expected_cols)


def col(name):
    return ArgumentSpec(True, name)



from pytest import raises

from qpd.specs import (
    AggFunctionSpec,
    ArgumentSpec,
    IsInSpec,
    IsValueSpec,
    NoWindowFrame,
    OrderBySpec,
    OrderItemSpec,
    RollingWindowFrame,
    WindowFunctionSpec,
    WindowSpec,
    make_windowframe_spec,
)


def test_argument_spec():
    a = ArgumentSpec(True, "x")
    assert a.is_col
    assert "x" == a.value


def test_agg_function_spec():
    f = AggFunctionSpec("max", unique=False, dropna=True)
    assert "max" == f.name
    assert not f.unique
    assert f.dropna
    print(f)
    assert f == f
    assert f != AggFunctionSpec("max", unique=True, dropna=True)
    s = set([f, f])
    assert len(s) == 1
    assert f in s
    assert AggFunctionSpec("max", unique=True, dropna=True) not in s


def test_order_item_spec():
    o = OrderItemSpec("x")
    assert o.asc
    assert "auto" == o.na_position
    assert "first" == o.pd_na_position
    o = OrderItemSpec("x", False)
    assert not o.asc
    assert "auto" == o.na_position
    assert "last" == o.pd_na_position
    o = OrderItemSpec("x", False, na_position="first")
    assert not o.asc
    assert "first" == o.na_position
    assert "first" == o.pd_na_position
    o = OrderItemSpec("x", True, na_position="last")
    assert o.asc
    assert "last" == o.na_position
    assert "last" == o.pd_na_position
    raises(ValueError, lambda: OrderItemSpec("x", True, na_position="x"))


def test_order_by_spec():
    a = OrderItemSpec("a")
    b = OrderItemSpec("b", False)
    c = OrderItemSpec("a", False, "last")
    d = OrderItemSpec("b", True, "first")
    e = OrderItemSpec("a", True, "last")
    f = OrderItemSpec("b", False, "first")
    o = OrderBySpec(a, b)
    assert len(o) == 2
    assert ["a", "b"] == o.keys
    assert [True, False] == o.asc
    assert "first" == o.pd_na_position
    o = OrderBySpec(b, e)
    assert "last" == o.pd_na_position
    o = OrderBySpec(a, f)
    assert "first" == o.pd_na_position
    raises(ValueError, lambda: OrderBySpec(c, d, e))
    assert "first" == OrderBySpec(a).pd_na_position
    assert "last" == OrderBySpec(b).pd_na_position
    assert "first" == OrderBySpec("a").pd_na_position
    assert "last" == OrderBySpec(("b", False)).pd_na_position
    assert "first" == OrderBySpec(("b", False, "first")).pd_na_position


def test_make_windowframe_spec():
    assert isinstance(make_windowframe_spec(""), NoWindowFrame)
    assert isinstance(make_windowframe_spec("rows", None, None), NoWindowFrame)
    assert isinstance(make_windowframe_spec("rows", None, 0), RollingWindowFrame)
    assert isinstance(make_windowframe_spec("rows", None, "0"), RollingWindowFrame)
    assert isinstance(make_windowframe_spec("rows", -1, "2"), RollingWindowFrame)
    w = make_windowframe_spec("rows", None, -5)
    assert w.start is None
    assert -5 == w.end
    assert w.window is None
    w = make_windowframe_spec("rows", "-6", 5)
    assert isinstance(w, RollingWindowFrame)
    assert -6 == w.start
    assert 5 == w.end
    assert 12 == w.window
    raises(NotImplementedError, lambda: make_windowframe_spec("x"))


def test_window_spec():
    wf = make_windowframe_spec("rows", -2, 0)
    w = WindowSpec("", ["a", "b"], OrderBySpec("c", ("d", False)), wf)
    assert w.name != ""
    assert ["a", "b"] == w.partition_keys
    assert ["c", "d"] == w.order_by.keys
    assert wf == w.windowframe
    w2 = WindowSpec("xy", ["a", "b"], OrderBySpec("c", ("d", False)), wf)
    assert w == w and w != w2
    assert "xy" == w2.name


def test_window_function():
    wf = make_windowframe_spec("rows", -2, 0)
    w = WindowSpec("", ["a", "b"], [("c", True, True), ("d", False, True)], wf)
    f = WindowFunctionSpec("sum", unique=True, dropna=False, window=w)
    assert f.window is w
    assert f.has_windowframe
    assert f.has_partition_by
    assert f.has_order_by


def test_is():
    v = IsValueSpec("NULL", False)
    assert "null" == v.name
    assert not v.positive
    assert "null" == v.value_expr

    v = IsInSpec("beTween", False)
    assert "between" == v.name
    assert v.is_between
    assert not v.positive
    assert 0 == len(v.values)

    a1 = ArgumentSpec(False, 1)
    a2 = ArgumentSpec(False, 2)
    v = IsInSpec("beTween", False, a1, a2)
    assert a1 is v.values[0]
    assert a2 is v.values[1]



from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
from pandas.core.window.indexers import BaseIndexer
from qpd import PandasLikeUtils, QPDEngine, run_sql
from qpd.dataframe import Column, DataFrame
from qpd.specs import (
    AggFunctionSpec,
    ArgumentSpec,
    OrderBySpec,
    RollingWindowFrame,
    WindowFunctionSpec,
)
from triad.utils.assertion import assert_or_throw


def run_sql_on_pandas(
    sql: str, dfs: Dict[str, Any], ignore_case: bool = False
) -> pd.DataFrame:
    return run_sql(QPDPandasEngine(), sql, dfs, ignore_case=ignore_case)


class PandasUtils(PandasLikeUtils[pd.DataFrame, pd.Series]):
    pass


class QPDPandasEngine(QPDEngine):
    @property
    def pl_utils(self) -> PandasUtils:
        return PandasUtils()

    def to_df(self, obj: Any) -> DataFrame:
        if isinstance(obj, DataFrame):
            return obj
        if isinstance(obj, pd.DataFrame):
            return DataFrame(Column(v, k) for k, v in obj.to_dict("series").items())
        raise ValueError(f"{obj} is not supported")  # pragma: no cover

    def to_col(self, value: Any, name: str = "") -> Column:
        if isinstance(value, list):
            return Column(pd.Series(value), name)
        return Column(value, name)

    def to_native(self, df: DataFrame) -> Any:
        if all(not self.is_series(x.native) for x in df.values()):
            return pd.DataFrame({k: [v.native] for k, v in df.items()})
        return pd.DataFrame({k: v.native for k, v in df.items()})

    def is_series(self, obj: Any) -> bool:
        return isinstance(obj, pd.Series)

    def case_when(self, *cols: Column) -> Column:
        """`cols` must be in the format of
        `when1`, `value1`, ... ,`default`, and length must be >= 3.
        The reason to design the interface in this way is to simplify the translation
        from SQL to engine APIs.

        :return: [description]
        :rtype: Column
        """
        assert_or_throw(len(cols) >= 3, "case when must have 1+ conditions")

        def where(pos: int) -> Any:
            return np.where(
                cols[pos].native,
                cols[pos + 1].native,
                cols[pos + 2].native if pos + 3 == len(cols) else where(pos + 2),
            )

        return Column(pd.Series(where(0)))

    def order_by_limit(
        self, df: DataFrame, order_by: OrderBySpec, limit: int
    ) -> DataFrame:
        if len(order_by) == 0 and limit < 0:  # pragma: no cover
            return df
        ndf = self.to_native(df)
        if len(order_by) > 0:
            ndf = self._safe_sort(ndf, order_by)
        if limit >= 0:
            ndf = ndf.head(limit)
        return self.to_df(ndf)

    def group_agg(
        self,
        df: DataFrame,
        keys: List[str],
        agg_map: Dict[str, Tuple[str, AggFunctionSpec]],
    ) -> DataFrame:
        if len(keys) > 0:
            return self._group_agg_with_keys(df, keys, agg_map)
        else:
            return self._group_agg_no_keys(df, agg_map)

    def window(  # noqa: C901
        self,
        df: DataFrame,
        func: WindowFunctionSpec,
        args: List[ArgumentSpec],
        dest_col_name: str,
    ) -> DataFrame:
        if func.name == "row_number":
            return self._window_row_number(df, func, args, dest_col_name)
        if func.name == "sum":
            return self._window_simple_agg(
                lambda x: x.sum(min_count=1), df, func, args, dest_col_name
            )
        if func.name == "min":
            return self._window_simple_agg(
                lambda x: x.min(), df, func, args, dest_col_name
            )
        if func.name == "max":
            return self._window_simple_agg(
                lambda x: x.max(), df, func, args, dest_col_name
            )
        if func.name == "count":
            return self._window_simple_agg(
                lambda x: x.count(), df, func, args, dest_col_name
            )
        if func.name in ["avg", "mean"]:
            return self._window_simple_agg(
                lambda x: x.mean(), df, func, args, dest_col_name
            )
        if func.name in ["rank"]:
            return self._window_ranks(
                lambda x: x.rank(method="min"), df, func, args, dest_col_name
            )
        if func.name in ["dense_rank"]:
            return self._window_ranks(
                lambda x: x.rank(method="dense"), df, func, args, dest_col_name
            )
        if func.name in ["percentile_rank", "percent_rank"]:

            def pct_rank(s: Any) -> Any:
                m = s.size - 1
                ranks = s.rank(method="min") - 1
                if m == 0:
                    m = 1.0
                return ranks / m

            return self._window_ranks(
                lambda x: pct_rank(x)
                if isinstance(x, pd.Series)
                else x.apply(pct_rank),
                df,
                func,
                args,
                dest_col_name,
            )
        if func.name == "lead":
            return self._window_lead_lag(
                lambda x, delta, default: x.shift(-delta, fill_value=default),
                df,
                func,
                args,
                dest_col_name,
            )
        if func.name == "lag":
            return self._window_lead_lag(
                lambda x, delta, default: x.shift(delta, fill_value=default),
                df,
                func,
                args,
                dest_col_name,
            )
        raise NotImplementedError(f"{func}")  # pragma: no cover

    def _window_row_number(
        self,
        df: DataFrame,
        func: WindowFunctionSpec,
        args: List[ArgumentSpec],
        dest_col_name: str,
    ) -> DataFrame:
        assert_or_throw(len(args) == 0, ValueError(f"{args}"))
        assert_or_throw(
            not func.has_windowframe, ValueError("row_number can't have windowframe")
        )
        ndf = self.to_native(df)
        ndf = self._safe_sort(ndf, func.window.order_by)
        if func.has_partition_by > 0:
            gp, _ = self._safe_groupby(ndf, func.window.partition_keys)
            col = gp.cumcount() + 1
        else:
            col = np.arange(1, 1 + ndf.shape[0])
        ndf[dest_col_name] = col
        return self.to_df(ndf)

    def _window_ranks(
        self,
        rank_func: Callable,
        df: DataFrame,
        func: WindowFunctionSpec,
        args: List[ArgumentSpec],
        dest_col_name: str,
    ) -> DataFrame:
        assert_or_throw(len(args) == 0, ValueError(f"{args}"))
        assert_or_throw(
            func.has_order_by, ValueError(f"rank functions require order by {func}")
        )
        assert_or_throw(
            not func.has_windowframe,
            ValueError(f"rank functions can't have windowframe {func}"),
        )
        ndf = self.to_native(df)
        ndf = self._safe_sort(ndf, func.window.order_by)
        rank_on = self._safe_groupby(
            ndf, func.window.order_by.keys, as_index=False, sort=False
        )[0].ngroup()
        ndf["_rank_on"] = rank_on
        if len(func.window.partition_keys) > 0:
            gp, _ = self._safe_groupby(ndf, func.window.partition_keys)
            col = rank_func(gp["_rank_on"])
        else:
            col = rank_func(ndf["_rank_on"])
        ndf[dest_col_name] = col
        ndf = ndf[list(df.keys()) + [dest_col_name]]
        return self.to_df(ndf)

    def _window_lead_lag(
        self,
        l_func: Callable,
        df: DataFrame,
        func: WindowFunctionSpec,
        args: List[ArgumentSpec],
        dest_col_name: str,
    ) -> DataFrame:
        assert_or_throw(len(args) in [2, 3], ValueError(f"{args}"))
        assert_or_throw(args[0].is_col, ValueError(f"{args[0]}"))
        assert_or_throw(
            not args[1].is_col and isinstance(args[1].value, int),
            ValueError(f"{args[1]}"),
        )
        if len(args) == 3:
            assert_or_throw(not args[2].is_col, ValueError(f"{args[2]}"))
        assert_or_throw(
            not func.has_windowframe,
            ValueError(f"lead/lag functions can't have windowframe {func}"),
        )
        col = args[0].value
        delta = int(args[1].value)
        default = None if len(args) == 2 else args[2].value
        ndf = self.to_native(df)
        ndf = self._safe_sort(ndf, func.window.order_by)
        if len(func.window.partition_keys) > 0:
            gp, _ = self._safe_groupby(ndf, func.window.partition_keys)
            col = l_func(gp[col], delta, default)
        else:
            col = l_func(ndf[col], delta, default)
        ndf[dest_col_name] = col
        ndf = ndf[list(df.keys()) + [dest_col_name]]
        return self.to_df(ndf)

    def _window_simple_agg(  # noqa: C901
        self,
        agg_func: Callable,
        df: DataFrame,
        func: WindowFunctionSpec,
        args: List[ArgumentSpec],
        dest_col_name: str,
    ) -> DataFrame:
        assert_or_throw(len(args) == 1 and args[0].is_col, ValueError(f"{args}"))
        assert_or_throw(
            not func.unique and not func.dropna,
            NotImplementedError(
                f"window function doesn't support unique {func.unique} "
                + f"and dropna {func.dropna}"
            ),
        )
        ndf = self.to_native(df)
        ndf = self._safe_sort(ndf, func.window.order_by)
        col = args[0].value
        if not func.has_windowframe:
            if not func.has_partition_by:
                res: Any = agg_func(ndf[col])
            else:
                gp, _ = self._safe_groupby(ndf, func.window.partition_keys)
                res = gp[col].transform(agg_func)
        else:
            wf = func.window.windowframe
            if isinstance(wf, RollingWindowFrame):
                min_p = 0 if func.name in ["count", "min", "max"] else 1
                # 1.1.0 has this bug https://github.com/pandas-dev/pandas/issues/35557
                # we try to use native functions here to bypass bugs
                if wf.start is None and wf.end == 0:
                    rolling_func: Callable = lambda x: x.expanding(min_periods=min_p)
                elif wf.window is not None and wf.end == 0:
                    rolling_func: Callable = lambda x: x.rolling(  # type: ignore
                        window=wf.window, min_periods=min_p
                    )
                else:  # will not work on version pandas 1.1.0
                    rolling_func: Callable = lambda x: x.rolling(  # type: ignore
                        window=_RowsIndexer(wf.start, wf.end), min_periods=min_p
                    )
                if not func.has_partition_by:
                    res = ndf[col]
                    res = agg_func(rolling_func(res))
                else:
                    gp, _ = self._safe_groupby(
                        ndf, func.window.partition_keys, as_index=True
                    )
                    res = agg_func(rolling_func(gp[col]))
                    levels = list(range(len(res.index.names) - 1))
                    if len(levels) > 0:
                        res = res.reset_index(level=levels, drop=True)
            else:  # pragma: no cover
                raise NotImplementedError(wf)
        ndf[dest_col_name] = res
        return self.to_df(ndf)

    def _group_agg_with_keys(
        self,
        df: DataFrame,
        keys: List[str],
        agg_map: Dict[str, Tuple[str, AggFunctionSpec]],
    ) -> DataFrame:
        ndf = self.to_native(df)
        simple_agg_map = {
            k: pd.NamedAgg(column=v[0], aggfunc=self._map_agg_function(v[1]))
            for k, v in agg_map.items()
            if "*" not in v[0] and "," not in v[0]
        }
        gp, gp_keys = self._safe_groupby(ndf, keys, keep_extra_keys=True)
        if len(simple_agg_map) > 0:
            res = gp.agg(**simple_agg_map)
        cts: Dict[str, Any] = {}
        for k, v in agg_map.items():
            if "*" in v[0] or "," in v[0]:
                if v[1].unique:
                    cts[k] = self._count_unique(ndf, gp_keys, v[0], v[1])
                else:
                    cts[k] = self._count_all(gp, v[1])
        if len(simple_agg_map) > 0 and len(cts) == 0:
            return self.to_df(res.reset_index(drop=True))
        if len(simple_agg_map) == 0 and len(cts) > 0:
            return self.to_df(pd.DataFrame(cts).reset_index(drop=True))
        if len(simple_agg_map) > 0 and len(cts) > 0:
            for k, vv in cts.items():
                res[k] = vv
            return self.to_df(res.reset_index(drop=True))
        raise NotImplementedError("Impossible case")  # pragma: no cover

    def _group_agg_no_keys(
        self,
        df: DataFrame,
        agg_map: Dict[str, Tuple[str, AggFunctionSpec]],
    ) -> DataFrame:
        ndf = self.to_native(df)
        simple_agg_map = {
            k: (v[0], self._map_agg_function_no_group(v[1]))
            for k, v in agg_map.items()
            if "*" not in v[0] and "," not in v[0]
        }
        aggs: Dict[str, Any] = {}
        for k, v in simple_agg_map.items():
            aggs[k] = ndf.agg({v[0]: [v[1]]}).iloc[0, 0]
        for k, v in agg_map.items():
            if "*" in v[0] or "," in v[0]:
                if v[1].unique:
                    aggs[k] = self._count_unique_no_group(ndf, v[0], v[1])
                else:
                    aggs[k] = self._count_all_no_group(ndf, v[1])
        res = pd.DataFrame({k: [v] for k, v in aggs.items()})
        return self.to_df(res)

    def _map_agg_function(self, func: AggFunctionSpec) -> Any:  # noqa: C901
        # TODO: this function has compatibility issue with pandas < 1
        # for example: we want to compute the sum of a same column twice
        # https://github.com/pandas-dev/pandas/issues/27519
        name = func.name.lower()
        if name == "sum":
            if not func.unique:
                return lambda x: x.sum(min_count=1)
            else:
                return lambda x: x.drop_duplicates().sum(min_count=1)
        if name in ["avg", "mean"]:
            if not func.unique:
                return "mean"
            else:
                return lambda x: x.drop_duplicates().mean()
        if name in ["first", "first_value"]:
            if func.dropna:
                return "first"
            else:
                return lambda x: x.head(1)
        if name in ["last", "last_value"]:
            if func.dropna:
                return "last"
            else:
                return lambda x: x.tail(1)
        if name == "count":
            if not func.unique and not func.dropna:
                return "size"
            if func.unique:
                return lambda x: x.nunique(dropna=func.dropna)
            if not func.unique and func.dropna:
                return "count"
        if name == "min":
            # probably related with https://github.com/pandas-dev/pandas/issues/31777
            # pandas has to be >=1.0.2
            return lambda x: x.dropna().min()
        if name == "max":
            return lambda x: x.dropna().max()
        return name  # pragma: no cover

    def _count_all(self, df_agg: Any, func: AggFunctionSpec) -> Any:
        assert_or_throw(
            "count" == func.name.lower() and not func.unique,
            ValueError(f"{func} is invalid"),
        )
        return df_agg.size()

    def _count_unique(
        self, df: Any, keys: List[str], cols: str, func: AggFunctionSpec
    ) -> Any:
        assert_or_throw(
            "count" == func.name.lower() and func.unique,
            ValueError(f"{func} can't be applied on {cols}"),
        )
        if cols == "*":
            dkeys = set(df.columns)
        else:
            dkeys = set(keys)
            dkeys.update(cols.split(","))
        return df[dkeys].drop_duplicates().groupby(keys).size()

    def _map_agg_function_no_group(self, func: AggFunctionSpec) -> Any:  # noqa: C901
        name = func.name.lower()
        if name == "sum":
            if not func.unique:
                return lambda x: x.sum(min_count=1)
            else:
                return lambda x: x.drop_duplicates().sum(min_count=1)
        if name in ["avg", "mean"]:
            if not func.unique:
                return "mean"
            else:
                return lambda x: x.drop_duplicates().mean()
        if name in ["first", "first_value"]:
            if func.dropna:
                return lambda x: x[x.first_valid_index()]
            else:
                return lambda x: x.head(1)
        if name in ["last", "last_value"]:
            if func.dropna:
                return lambda x: x[x.last_valid_index()]
            else:
                return lambda x: x.tail(1)
        if name == "count":
            if not func.unique and not func.dropna:
                return "size"
            if func.unique and not func.dropna:
                return lambda x: x.drop_duplicates().size
            if func.unique and func.dropna:
                return "nunique"
            if not func.unique and func.dropna:
                return "count"
        if name == "min":
            return lambda x: x.dropna().min()
        if name == "max":
            return lambda x: x.dropna().max()
        return name  # pragma: no cover

    def _count_all_no_group(self, df: Any, func: AggFunctionSpec) -> Any:
        assert_or_throw(
            "count" == func.name.lower() and not func.unique,
            ValueError(f"{func} is invalid"),
        )
        return df.shape[0]

    def _count_unique_no_group(self, df: Any, cols: str, func: AggFunctionSpec) -> Any:
        assert_or_throw(
            "count" == func.name.lower() and func.unique,
            ValueError(f"{func} can't be applied on {cols}"),
        )
        if cols == "*":
            dkeys = set(df.columns)
        else:
            dkeys = set(cols.split(","))
        return df[dkeys].drop_duplicates().shape[0]

    def _safe_groupby(
        self,
        ndf: Any,
        keys: List[str],
        keep_extra_keys: bool = False,
        as_index: bool = True,
        sort: bool = True,
    ) -> Tuple[Any, List[str]]:
        if not keep_extra_keys:
            ndf = ndf[ndf.columns]
        nulldf = ndf[keys].isnull()
        gp_keys: List[str] = []
        for k in keys:
            ndf[k + "_n"] = nulldf[k]
            ndf[k + "_g"] = ndf[k].fillna(0)
            gp_keys.append(k + "_n")
            gp_keys.append(k + "_g")
        return ndf.groupby(gp_keys, as_index=as_index, sort=sort), gp_keys

    def _safe_sort(self, ndf: Any, order_by: OrderBySpec) -> Any:
        if len(order_by.keys) == 0:
            return ndf
        keys = list(ndf.columns)
        okeys: List[str] = []
        asc: List[bool] = []
        for oi in order_by:
            nk = oi.name + "_null"
            ndf[nk] = ndf[oi.name].isnull()
            okeys.append(nk)
            asc.append(oi.pd_na_position != "first")
            okeys.append(oi.name)
            asc.append(oi.asc)
        ndf = ndf.sort_values(okeys, ascending=asc)
        return ndf[keys]


class _RowsIndexer(BaseIndexer):
    def __init__(self, start: Optional[int], end: Optional[int]):
        window = 0 if start is None or end is None else end - start + 1
        super().__init__(self, window_size=window, start=start, end=end)

    def get_window_bounds(
        self,
        num_values: int = 0,
        min_periods: Optional[int] = None,
        center: Optional[bool] = None,
        closed: Optional[str] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        if self.start is None:
            start = np.zeros(num_values, dtype=np.int64)
        else:
            start = np.arange(self.start, self.start + num_values, dtype=np.int64)
            if self.start < 0:
                start[: -self.start] = 0
            elif self.start > 0:
                start[-self.start :] = num_values
        if self.end is None:
            end = np.full(num_values, num_values, dtype=np.int64)
        else:
            end = np.arange(self.end + 1, self.end + 1 + num_values, dtype=np.int64)
            if self.end > 0:
                end[-self.end :] = num_values
            elif self.end < 0:
                end[: -self.end] = 0
        return start, end



from typing import Any

import pandas as pd

from qpd.dataframe import Column, DataFrame
from qpd_pandas import QPDPandasEngine
from qpd_pandas.engine import _RowsIndexer
from qpd_test.engine_suite import QPDEngineTests


class EngineTests(QPDEngineTests.Tests):
    def to_pandas_df(self, df: Any) -> pd.DataFrame:
        if isinstance(df, DataFrame):
            return self.op.to_native(df)
        elif isinstance(df, pd.DataFrame):
            return df
        else:
            raise ValueError(df)

    def to_pandas_series(self, col: Column) -> pd.Series:
        return col.native

    def make_qpd_engine(self):
        return QPDPandasEngine()

    def to_native_df(self, data: Any, columns: Any) -> Any:  # pragma: no cover
        if isinstance(data, pd.DataFrame):
            return data
        return pd.DataFrame(data, columns=columns)


def test_rows_indexer():
    r = _RowsIndexer(None, None)
    a, b = r.get_window_bounds(0)
    assert [] == a.tolist()
    assert [] == b.tolist()
    a, b = r.get_window_bounds(2)
    assert [0, 0] == a.tolist()
    assert [2, 2] == b.tolist()

    r = _RowsIndexer(None, -2)
    a, b = r.get_window_bounds(0)
    assert [] == a.tolist()
    assert [] == b.tolist()
    a, b = r.get_window_bounds(3)
    assert [0, 0, 0] == a.tolist()
    assert [0, 0, 1] == b.tolist()

    r = _RowsIndexer(None, 0)
    a, b = r.get_window_bounds(0)
    assert [] == a.tolist()
    assert [] == b.tolist()
    a, b = r.get_window_bounds(3)
    assert [0, 0, 0] == a.tolist()
    assert [1, 2, 3] == b.tolist()

    r = _RowsIndexer(None, 1)
    a, b = r.get_window_bounds(0)
    assert [] == a.tolist()
    assert [] == b.tolist()
    a, b = r.get_window_bounds(3)
    assert [0, 0, 0] == a.tolist()
    assert [2, 3, 3] == b.tolist()

    r = _RowsIndexer(-1, None)
    a, b = r.get_window_bounds(0)
    assert [] == a.tolist()
    assert [] == b.tolist()
    a, b = r.get_window_bounds(3)
    assert [0, 0, 1] == a.tolist()
    assert [3, 3, 3] == b.tolist()

    r = _RowsIndexer(0, None)
    a, b = r.get_window_bounds(0)
    assert [] == a.tolist()
    assert [] == b.tolist()
    a, b = r.get_window_bounds(3)
    assert [0, 1, 2] == a.tolist()
    assert [3, 3, 3] == b.tolist()

    r = _RowsIndexer(2, None)
    a, b = r.get_window_bounds(0)
    assert [] == a.tolist()
    assert [] == b.tolist()
    a, b = r.get_window_bounds(3)
    assert [2, 3, 3] == a.tolist()
    assert [3, 3, 3] == b.tolist()

    r = _RowsIndexer(-3, -1)
    a, b = r.get_window_bounds(0)
    assert [] == a.tolist()
    assert [] == b.tolist()
    a, b = r.get_window_bounds(3)
    assert [0, 0, 0] == a.tolist()
    assert [0, 1, 2] == b.tolist()

    r = _RowsIndexer(-1, 0)
    a, b = r.get_window_bounds(0)
    assert [] == a.tolist()
    assert [] == b.tolist()
    a, b = r.get_window_bounds(3)
    assert [0, 0, 1] == a.tolist()
    assert [1, 2, 3] == b.tolist()

    r = _RowsIndexer(0, 1)
    a, b = r.get_window_bounds(0)
    assert [] == a.tolist()
    assert [] == b.tolist()
    a, b = r.get_window_bounds(3)
    assert [0, 1, 2] == a.tolist()
    assert [2, 3, 3] == b.tolist()

    r = _RowsIndexer(2, 3)
    a, b = r.get_window_bounds(0)
    assert [] == a.tolist()
    assert [] == b.tolist()
    a, b = r.get_window_bounds(3)
    assert [2, 3, 3] == a.tolist()
    assert [3, 3, 3] == b.tolist()



from typing import Any

import ray

ray.init(ignore_reinit_error=True)
import os

os.environ["MODIN_ENGINE"] = "ray"
import modin.pandas as pd
import pandas

from qpd.dataframe import Column, DataFrame
from qpd_ray import QPDRayEngine, run_sql_on_ray
from qpd_test.sql_suite import SQLTests
from qpd_test.utils import assert_df_eq


def test_entrypoint():
    pdf = pd.DataFrame([[0, 1], [1, 2], [0, 4]], columns=["a", "b"])
    df1 = run_sql_on_ray(
        "SELECT a, SUM(b) AS b, COUNT(*) AS c FROM df GROUP BY a", {"df": pdf}
    )
    df2 = run_sql_on_ray("select a+1 as a, b from df", {"df": pdf}, ignore_case=True)
    res = run_sql_on_ray(
        """
        SELECT a.a, a.b, c, b.b AS d 
        FROM a INNER JOIN b ON a.a=b.a
        """,
        dict(a=df1, b=df2),
    )
    assert_df_eq(res._to_pandas(), [[1, 2, 1, 1], [1, 2, 1, 4]], list("abcd"))


class SQLTests(SQLTests.Tests):
    def test_agg_first_last(self):
        a = ([[1, "x", None], [2, None, 3.5]], ["a", "b", "c"])
        self.assert_eq(
            dict(a=a),
            """
            SELECT
                a,
                FIRST(b) AS a1,
                LAST(b) AS a2,
                FIRST_VALUE(c) AS a3,
                LAST_VALUE(c) AS a4
            FROM a GROUP BY a
            """,
            [[1, "x", "x", None, None], [2, None, None, 3.5, 3.5]],
            ["a", "a1", "a2", "a3", "a4"],
        )

    def test_agg_first_last_no_group_by(self):
        return

    def test_order_by_no_limit(self):
        return

    def test_window_count(self):
        return

    def test_window_lead_lag(self):
        return

    def test_window_min_max(self):
        return

    def test_window_ranks(self):
        return

    def test_window_row_number(self):
        return

    def test_window_sum_avg(self):
        return

    def to_pandas_df(self, df: Any) -> pandas.DataFrame:
        if isinstance(df, DataFrame):
            return self.op.to_native(df)._to_pandas()
        elif isinstance(df, pd.DataFrame):
            return df._to_pandas()
        elif isinstance(df, pandas.DataFrame):
            return df
        else:
            raise ValueError(df)

    def to_pandas_series(self, col: Column) -> pandas.Series:
        return pandas.Series(col.native._to_pandas(), name=col.name)

    def make_qpd_engine(self):
        return QPDRayEngine()

    def to_native_df(self, data: Any, columns: Any) -> Any:
        if isinstance(data, pd.DataFrame):
            return data
        if isinstance(data, pandas.DataFrame):
            return pd.DataFrame(data)
        return pd.DataFrame(data, columns=columns)
